#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import jax.numpy as jnp
import jax.random as jr

from phydrax.domain import Interval1d, ProductStructure, TimeInterval


def test_time_only_function_broadcasts_over_space_and_time_axes():
    geom = Interval1d(0.0, 1.0)
    time = TimeInterval(0.0, 1.0)
    dom = geom @ time

    @dom.Function("t")
    def f(t):
        return 3.0 * t

    component = dom.component()
    structure = ProductStructure((("x",), ("t",)))
    batch = component.sample((7, 9), structure=structure, key=jr.key(0))

    out = f(batch)

    axis_x = batch.structure.axis_for("x")
    axis_t = batch.structure.axis_for("t")
    assert axis_x is not None and axis_t is not None
    assert set(out.named_dims) == {axis_x, axis_t}

    t_vals = jnp.asarray(batch["t"].data, dtype=float).reshape((-1,))
    n_x = int(batch["x"].data.shape[0])
    if out.dims == (axis_t, axis_x):
        expected = jnp.broadcast_to(3.0 * t_vals[:, None], (t_vals.shape[0], n_x))
    elif out.dims == (axis_x, axis_t):
        expected = jnp.broadcast_to(3.0 * t_vals[None, :], (n_x, t_vals.shape[0]))
    else:
        raise AssertionError(f"Unexpected dims for broadcasted output: {out.dims}")

    assert jnp.allclose(out.data, expected)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

from typing import cast

import coordax as cx
import equinox as eqx
import jax.numpy as jnp
import jax.random as jr

from phydrax.domain import (
    DomainFunction,
    Interval1d,
    ProductStructure,
    Square,
    TimeInterval,
)
from phydrax.nn.models import LatentContractionModel
from phydrax.nn.models.core._base import _AbstractBaseModel
from phydrax.operators.differential import partial_t, partial_x, partial_y
from phydrax.operators.integral import integral


def _as_scalar(x):
    arr = jnp.asarray(x)
    if arr.ndim == 0:
        return arr
    if arr.size != 1:
        raise ValueError("Expected scalar input for scalar factor model.")
    return arr.reshape(())


class XYLatentModel(_AbstractBaseModel):
    in_size: int | str
    out_size: int | str

    def __init__(self) -> None:
        self.in_size = 2
        self.out_size = 2

    def __call__(self, x, /, *, key=jr.key(0)):
        x = jnp.asarray(x)
        return jnp.stack([x[0] + x[1], jnp.array(1.0)], axis=-1)


class ScalarLatentModel(_AbstractBaseModel):
    in_size: int | str
    out_size: int | str

    def __init__(self) -> None:
        self.in_size = "scalar"
        self.out_size = 2

    def __call__(self, x, /, *, key=jr.key(0)):
        x = _as_scalar(x)
        return jnp.stack([x, jnp.array(1.0)], axis=-1)


class ConstantXYLatentModel(_AbstractBaseModel):
    in_size: int | str
    out_size: int | str

    def __init__(self) -> None:
        self.in_size = 2
        self.out_size = 1

    def __call__(self, x, /, *, key=jr.key(0)):
        return jnp.array([1.0])


class ScalarOffsetLatentModel(_AbstractBaseModel):
    in_size: int | str
    out_size: int | str
    offset: float

    def __init__(self, offset: float) -> None:
        self.in_size = "scalar"
        self.out_size = 1
        self.offset = float(offset)

    def __call__(self, x, /, *, key=jr.key(0)):
        x = _as_scalar(x)
        return jnp.array([x + self.offset])


def _latent_function(domain, model):
    def f(x, p, t, *, key=jr.key(0)):
        return model({"x": x, "p": p, "t": t}, key=key)

    return DomainFunction(domain=domain, deps=("x", "p", "t"), func=f)


def _squeeze_field_for_compare(
    field: cx.Field,
) -> tuple[jnp.ndarray, tuple[str | None, ...]]:
    data = jnp.asarray(field.data)
    dims = list(field.dims)
    squeeze_axes = [
        i
        for i, (d, n) in enumerate(zip(dims, data.shape, strict=True))
        if d is None and n == 1
    ]
    if squeeze_axes:
        data = jnp.squeeze(data, axis=tuple(squeeze_axes))
        dims = [d for i, d in enumerate(dims) if i not in squeeze_axes]
    return data, tuple(dims)


def _assert_field_allclose(field, expected, *, atol=1e-6):
    lhs, lhs_dims = _squeeze_field_for_compare(field)
    rhs, rhs_dims = _squeeze_field_for_compare(expected)
    assert lhs_dims == rhs_dims
    assert jnp.allclose(lhs, rhs, atol=atol)


def _assert_array_allclose(actual, expected, *, atol=1e-6):
    a = jnp.asarray(actual)
    b = jnp.asarray(expected)
    if a.ndim > 0 and a.shape[-1] == 1:
        a = a[..., 0]
    if b.ndim > 0 and b.shape[-1] == 1:
        b = b[..., 0]
    assert jnp.allclose(a, b, atol=atol)


def _scalar_field_from_dense(field: cx.Field, axis: str) -> cx.Field:
    data = jnp.asarray(field.data)
    if data.ndim == 2 and data.shape[1] == 1:
        data = data[:, 0]
    if data.ndim != 1:
        raise ValueError(
            f"Expected dense scalar field with shape (N,) or (N,1); got {data.shape}."
        )
    return cx.Field(data, dims=(axis,))


def test_latent_contraction_product_domain_partials():
    geom = Square(center=(0.0, 0.0), side=2.0)
    param = Interval1d(-1.0, 1.0).relabel("p")
    time = TimeInterval(0.0, 1.0)
    domain = geom @ param @ time

    model = LatentContractionModel(
        latent_size=2,
        out_size="scalar",
        x=XYLatentModel(),
        p=ScalarLatentModel(),
        t=ScalarLatentModel(),
    )
    u = _latent_function(domain, model)

    component = domain.component()
    sep = component.sample_coord_separable(
        {"x": (4, 3)},
        num_points=(5, 4),
        dense_structure=ProductStructure((("p",), ("t",))),
        key=jr.key(0),
    )

    p_axis = sep.dense_structure.axis_for("p")
    t_axis = sep.dense_structure.axis_for("t")
    x_axes = sep.coord_axes_by_label["x"]

    p_field = _scalar_field_from_dense(sep.points["p"], p_axis)
    t_field = _scalar_field_from_dense(sep.points["t"], t_axis)
    x0_data = jnp.asarray(sep.points["x"][0].data)
    x1_data = jnp.asarray(sep.points["x"][1].data)
    x0 = cx.Field(x0_data, dims=(x_axes[0],))
    x1 = cx.Field(x1_data, dims=(x_axes[1],))
    x_sum = x0 + x1
    one_x0 = cx.Field(jnp.ones_like(x0_data), dims=(x_axes[0],))
    one_x1 = cx.Field(jnp.ones_like(x1_data), dims=(x_axes[1],))
    p_data = jnp.asarray(p_field.data)
    t_data = jnp.asarray(t_field.data)
    one_p = cx.Field(jnp.ones_like(p_data), dims=(p_axis,))
    one_t = cx.Field(jnp.ones_like(t_data), dims=(t_axis,))

    expected_u = (p_field * t_field) * x_sum + 1.0
    expected_dx = (p_field * t_field) * one_x0 * one_x1
    expected_dy = expected_dx
    expected_dp = (one_p * t_field) * x_sum
    expected_dt = (p_field * one_t) * x_sum

    _assert_field_allclose(u(sep), expected_u)

    du_dx = partial_x(u, var="x")
    du_dy = partial_y(u, var="x")
    du_dp = partial_x(u, var="p")
    du_dt = partial_t(u, var="t")

    eval_jit = eqx.filter_jit(lambda f, b: f(b).data)
    _assert_array_allclose(eval_jit(du_dx, sep), expected_dx.data)
    _assert_array_allclose(eval_jit(du_dy, sep), expected_dy.data)
    _assert_array_allclose(eval_jit(du_dp, sep), expected_dp.data)
    _assert_array_allclose(eval_jit(du_dt, sep), expected_dt.data)


def test_latent_contraction_product_domain_integral_over_x():
    geom = Square(center=(0.0, 0.0), side=2.0)
    param = Interval1d(-1.0, 1.0).relabel("p")
    time = TimeInterval(0.0, 1.0)
    domain = geom @ param @ time

    model = LatentContractionModel(
        latent_size=1,
        out_size="scalar",
        x=ConstantXYLatentModel(),
        p=ScalarOffsetLatentModel(1.0),
        t=ScalarOffsetLatentModel(2.0),
    )
    u = _latent_function(domain, model)

    component = domain.component()
    sep = component.sample_coord_separable(
        {"x": (6, 5)},
        num_points=(4, 3),
        dense_structure=ProductStructure((("p",), ("t",))),
        key=jr.key(1),
    )

    out = integral(u, sep, component=component, over="x")
    p_axis = sep.dense_structure.axis_for("p")
    t_axis = sep.dense_structure.axis_for("t")
    p_field = _scalar_field_from_dense(sep.points["p"], p_axis)
    t_field = _scalar_field_from_dense(sep.points["t"], t_axis)
    area = jnp.asarray(geom.volume, dtype=float)
    param_base = cast(Interval1d, param.base)
    p_weight = jnp.asarray(param_base.volume, dtype=float) / float(
        sep.points["p"].data.shape[0]
    )
    t_weight = jnp.asarray(time.measure, dtype=float) / float(
        sep.points["t"].data.shape[0]
    )
    expected = area * p_weight * t_weight * (p_field + 1.0) * (t_field + 2.0)
    _assert_field_allclose(out, expected)

    eval_jit = eqx.filter_jit(
        lambda b: integral(u, b, component=component, over="x").data
    )
    _assert_array_allclose(eval_jit(sep), expected.data)


def test_latent_contraction_product_domain_paired_dense_block():
    geom = Square(center=(0.0, 0.0), side=2.0)
    param = Interval1d(-1.0, 1.0).relabel("p")
    time = TimeInterval(0.0, 1.0)
    domain = geom @ param @ time

    model = LatentContractionModel(
        latent_size=2,
        out_size="scalar",
        x=XYLatentModel(),
        p=ScalarLatentModel(),
        t=ScalarLatentModel(),
    )
    u = _latent_function(domain, model)

    component = domain.component()
    sep = component.sample_coord_separable(
        {"x": (3, 2)},
        num_points=6,
        dense_structure=ProductStructure((("p", "t"),)),
        key=jr.key(2),
    )

    axis = sep.dense_structure.axis_for("p")
    x_axes = sep.coord_axes_by_label["x"]
    p_field = _scalar_field_from_dense(sep.points["p"], axis)
    t_field = _scalar_field_from_dense(sep.points["t"], axis)
    x0 = cx.Field(sep.points["x"][0].data, dims=(x_axes[0],))
    x1 = cx.Field(sep.points["x"][1].data, dims=(x_axes[1],))
    expected = (p_field * t_field) * (x0 + x1) + 1.0

    _assert_field_allclose(u(sep), expected)
    eval_jit = eqx.filter_jit(lambda b: u(b).data)
    _assert_array_allclose(eval_jit(sep), expected.data)


def test_latent_contraction_multi_coord_separable_labels():
    geom = Square(center=(0.0, 0.0), side=2.0)
    param = Interval1d(-1.0, 1.0).relabel("p")
    time = TimeInterval(0.0, 1.0)
    domain = geom @ param @ time

    model = LatentContractionModel(
        latent_size=2,
        out_size="scalar",
        x=XYLatentModel(),
        p=ScalarLatentModel(),
        t=ScalarLatentModel(),
    )
    u = _latent_function(domain, model)

    component = domain.component()
    sep = component.sample_coord_separable(
        {"x": (3, 2), "p": 4},
        num_points=5,
        dense_structure=ProductStructure((("t",),)),
        key=jr.key(3),
    )

    t_axis = sep.dense_structure.axis_for("t")
    x_axes = sep.coord_axes_by_label["x"]
    p_axis = sep.coord_axes_by_label["p"][0]
    t_field = cx.Field(sep.points["t"].data, dims=(t_axis,))
    x0 = cx.Field(sep.points["x"][0].data, dims=(x_axes[0],))
    x1 = cx.Field(sep.points["x"][1].data, dims=(x_axes[1],))
    p_field = sep.points["p"][0]
    expected = (t_field * (x0 + x1)) * p_field + 1.0

    _assert_field_allclose(u(sep), expected)
    eval_jit = eqx.filter_jit(lambda b: u(b).data)
    assert jnp.allclose(eval_jit(sep), expected.data, atol=1e-6)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr

from phydrax.nn.models import EquinoxModel, EquinoxStructuredModel


def test_equinox_model_value_layout_tensor_io():
    module = eqx.nn.MLP(in_size=4, out_size=6, width_size=8, depth=2, key=jr.key(0))
    model = EquinoxModel(module, in_size=(2, 2), out_size=(3, 2))

    x = jnp.ones((2, 2))
    y = model(x)
    assert y.shape == (3, 2)


def test_equinox_model_value_layout_scalar_io():
    module = eqx.nn.MLP(
        in_size="scalar",
        out_size="scalar",
        width_size=8,
        depth=2,
        key=jr.key(1),
    )
    model = EquinoxModel(module, in_size="scalar", out_size="scalar")

    x = jnp.array(1.25)
    y = model(x)
    assert y.shape == ()


def test_equinox_model_passthrough_forwards_key_and_kwargs():
    module = eqx.nn.Dropout(p=0.5, inference=False)
    model = EquinoxModel(module, in_size=4, out_size=4, layout="passthrough")

    x = jnp.ones((4,))
    y = model(x, key=jr.key(2), inference=False)
    assert y.shape == x.shape


def test_equinox_structured_model_passthrough_tuple_input():
    class TupleModule(eqx.Module):
        def __call__(self, x, *, key=None):
            del key
            a, b = x
            return a + 2.0 * b

    model = EquinoxStructuredModel(
        TupleModule(), in_size="scalar", out_size="scalar", layout="passthrough"
    )
    y = model((jnp.array(1.0), jnp.array(2.0)))
    assert y.shape == ()
    assert jnp.isclose(y, 5.0)


def test_equinox_structured_model_value_layout_concatenates_tuple():
    module = eqx.nn.MLP(
        in_size=2, out_size="scalar", width_size=8, depth=2, key=jr.key(3)
    )
    model = EquinoxStructuredModel(module, in_size=2, out_size="scalar", layout="value")

    y = model((jnp.array(1.0), jnp.array(2.0)))
    assert y.shape == ()



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import coordax as cx
import equinox as eqx
import jax.numpy as jnp

from phydrax._frozendict import frozendict
from phydrax.constraints import enforce_dirichlet, FunctionalConstraint
from phydrax.domain import (
    Boundary,
    FixedStart,
    PointsBatch,
    ProductStructure,
    TimeInterval,
)
from phydrax.domain.geometry2d import Square
from phydrax.domain.geometry3d import Cube
from phydrax.operators.differential import grad
from phydrax.solver import (
    EnforcedConstraintPipelines,
    EnforcedInteriorData,
    SingleFieldEnforcedConstraint,
)


def _paired_batch(domain, xs, ts):
    structure = ProductStructure((("x", "t"),)).canonicalize(domain.labels)
    axis_names = structure.axis_names
    assert axis_names is not None
    axis = axis_names[0]
    points = frozendict(
        {
            "x": cx.Field(jnp.asarray(xs, dtype=float), dims=(axis, None)),
            "t": cx.Field(jnp.asarray(ts, dtype=float).reshape((-1,)), dims=(axis,)),
        }
    )
    return PointsBatch(points=points, structure=structure)


def _eval(domain, u_enforced, xs, ts):
    batch = _paired_batch(domain, xs=xs, ts=ts)
    return jnp.asarray(u_enforced(batch).data).reshape((-1,))


def test_mixed_constraints_2d_transient():
    geom = Square(center=(0.0, 0.0), side=2.0)
    time = TimeInterval(0.0, 1.0)
    domain = geom @ time

    @domain.Function("x", "t")
    def u(x, t):
        return x[0] + 2.0 * x[1] + t

    left = domain.component({"x": Boundary()}, where={"x": lambda p: p[0] < -0.9})
    right = domain.component({"x": Boundary()}, where={"x": lambda p: p[0] > 0.9})
    initial = domain.component({"t": FixedStart()})

    constraints = [
        SingleFieldEnforcedConstraint(
            "u",
            left,
            lambda f: enforce_dirichlet(f, left, var="x", target=1.0),
        ),
        SingleFieldEnforcedConstraint(
            "u",
            right,
            lambda f: enforce_dirichlet(f, right, var="x", target=2.0),
        ),
        SingleFieldEnforcedConstraint(
            "u",
            initial,
            lambda f: enforce_dirichlet(f, initial, var="t", target=3.0),
            time_derivative_order=0,
            initial_target=3.0,
        ),
        SingleFieldEnforcedConstraint(
            "u",
            initial,
            lambda f: enforce_dirichlet(f, initial, var="t", target=3.0),
            time_derivative_order=1,
            initial_target=1.0,
        ),
        SingleFieldEnforcedConstraint(
            "u",
            initial,
            lambda f: enforce_dirichlet(f, initial, var="t", target=3.0),
            time_derivative_order=2,
            initial_target=0.0,
        ),
    ]

    sensors = jnp.array([[0.2, 0.1], [0.4, -0.2]], dtype=float)
    times = jnp.array([0.25, 0.75], dtype=float)
    sensor_values = jnp.array([[4.0, 5.0], [6.0, 7.0]], dtype=float)
    interior = EnforcedInteriorData(
        "u",
        sensors=sensors,
        times=times,
        sensor_values=sensor_values,
    )

    pipelines = EnforcedConstraintPipelines.build(
        functions={"u": u},
        constraints=constraints,
        interior_data=[interior],
        num_reference=256,
    )
    u_enforced = pipelines.apply({"u": u})["u"]

    out = _eval(
        domain,
        u_enforced,
        xs=jnp.array([[-1.0, 0.0], [1.0, 0.0]]),
        ts=jnp.array([0.5, 0.5]),
    )
    assert jnp.allclose(out[0], 1.0, atol=5e-2)
    assert jnp.allclose(out[1], 2.0, atol=5e-2)

    out = _eval(domain, u_enforced, xs=jnp.array([[0.0, 0.0]]), ts=jnp.array([0.0]))
    assert jnp.allclose(out, 3.0, atol=2e-2)

    xs = jnp.repeat(sensors, times.shape[0], axis=0)
    ts = jnp.tile(times, sensors.shape[0])
    expected = sensor_values.reshape((-1,))
    out = _eval(domain, u_enforced, xs=xs, ts=ts)
    assert jnp.allclose(out, expected, atol=1e-3)

    soft = FunctionalConstraint.from_operator(
        component=domain.component({"x": Boundary()}),
        operator=lambda f: grad(f, var="x"),
        constraint_vars="u",
        num_points=16,
        structure=ProductStructure((("x", "t"),)),
        reduction="mean",
    )
    loss = eqx.filter_jit(lambda: soft.loss({"u": u_enforced}))()
    assert jnp.isfinite(loss)


def test_mixed_constraints_3d_transient():
    geom = Cube(center=(0.0, 0.0, 0.0), side=2.0)
    time = TimeInterval(0.0, 1.0)
    domain = geom @ time

    @domain.Function("x", "t")
    def u(x, t):
        return x[0] + x[1] + x[2] + t

    left = domain.component({"x": Boundary()}, where={"x": lambda p: p[0] < -0.9})
    right = domain.component({"x": Boundary()}, where={"x": lambda p: p[0] > 0.9})
    initial = domain.component({"t": FixedStart()})

    constraints = [
        SingleFieldEnforcedConstraint(
            "u",
            left,
            lambda f: enforce_dirichlet(f, left, var="x", target=1.0),
        ),
        SingleFieldEnforcedConstraint(
            "u",
            right,
            lambda f: enforce_dirichlet(f, right, var="x", target=2.0),
        ),
        SingleFieldEnforcedConstraint(
            "u",
            initial,
            lambda f: enforce_dirichlet(f, initial, var="t", target=3.0),
            time_derivative_order=0,
            initial_target=3.0,
        ),
        SingleFieldEnforcedConstraint(
            "u",
            initial,
            lambda f: enforce_dirichlet(f, initial, var="t", target=3.0),
            time_derivative_order=1,
            initial_target=1.0,
        ),
        SingleFieldEnforcedConstraint(
            "u",
            initial,
            lambda f: enforce_dirichlet(f, initial, var="t", target=3.0),
            time_derivative_order=2,
            initial_target=0.0,
        ),
    ]

    sensors = jnp.array([[0.2, 0.1, -0.1], [0.4, -0.2, 0.3]], dtype=float)
    times = jnp.array([0.25, 0.75], dtype=float)
    sensor_values = jnp.array([[4.0, 5.0], [6.0, 7.0]], dtype=float)
    interior = EnforcedInteriorData(
        "u",
        sensors=sensors,
        times=times,
        sensor_values=sensor_values,
    )

    pipelines = EnforcedConstraintPipelines.build(
        functions={"u": u},
        constraints=constraints,
        interior_data=[interior],
        num_reference=256,
    )
    u_enforced = pipelines.apply({"u": u})["u"]

    out = _eval(
        domain,
        u_enforced,
        xs=jnp.array([[-1.0, 0.0, 0.0], [1.0, 0.0, 0.0]]),
        ts=jnp.array([0.5, 0.5]),
    )
    assert jnp.allclose(out[0], 1.0, atol=5e-2)
    assert jnp.allclose(out[1], 2.0, atol=5e-2)

    out = _eval(domain, u_enforced, xs=jnp.array([[0.0, 0.0, 0.0]]), ts=jnp.array([0.0]))
    assert jnp.allclose(out, 3.0, atol=2e-2)

    xs = jnp.repeat(sensors, times.shape[0], axis=0)
    ts = jnp.tile(times, sensors.shape[0])
    expected = sensor_values.reshape((-1,))
    out = _eval(domain, u_enforced, xs=xs, ts=ts)
    assert jnp.allclose(out, expected, atol=1e-3)

    soft = FunctionalConstraint.from_operator(
        component=domain.component({"x": Boundary()}),
        operator=lambda f: grad(f, var="x"),
        constraint_vars="u",
        num_points=16,
        structure=ProductStructure((("x", "t"),)),
        reduction="mean",
    )
    loss = eqx.filter_jit(lambda: soft.loss({"u": u_enforced}))()
    assert jnp.isfinite(loss)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

# Ensure JAX uses 64-bit floats by default for numerical robustness
import jax


jax.config.update("jax_enable_x64", True)

from . import (
    constraints,
    domain,
    nn,
    operators,
    solver,
)


# Explicit re-exports for star import
__all__ = [
    "constraints",
    "domain",
    "nn",
    "operators",
    "solver",
]



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import jax.numpy as jnp

from phydrax.domain import Interval1d, Square, TimeInterval
from phydrax.operators.differential import (
    bilaplacian,
    div_diag_k_grad,
    dt_n,
    laplacian,
    partial_n,
)


def test_dt_n_second_derivative_matches_ad_and_closed_form():
    time = TimeInterval(0.0, 1.0)

    @time.Function("t")
    def u(t):
        return t**3

    d2_ad = dt_n(u, var="t", order=2, backend="ad")
    d2_jet = dt_n(u, var="t", order=2, backend="jet")

    t = jnp.asarray(0.37)
    expected = 6.0 * t
    assert jnp.allclose(d2_jet.func(t), expected, rtol=1e-6, atol=1e-6)
    assert jnp.allclose(d2_ad.func(t), expected, rtol=1e-6, atol=1e-6)


def test_partial_n_coord_separable_second_derivative_matches_closed_form():
    geom = Interval1d(0.0, 1.0)

    @geom.Function("x")
    def u(x):
        return x[0] ** 3

    d2_jet = partial_n(u, var="x", order=2, backend="jet")

    coords = jnp.linspace(0.0, 1.0, 8)
    out = d2_jet.func((coords,))
    assert jnp.allclose(out, 6.0 * coords, rtol=1e-6, atol=1e-6)


def test_laplacian_backend_jet_matches_ad():
    geom = Interval1d(0.0, 1.0)

    @geom.Function("x")
    def u(x):
        return x[0] ** 3

    lap_ad = laplacian(u, var="x", backend="ad")
    lap_jet = laplacian(u, var="x", backend="jet")

    x = jnp.asarray([0.2])
    expected = 6.0 * x[0]
    assert jnp.allclose(lap_ad.func(x), expected, rtol=1e-6, atol=1e-6)
    assert jnp.allclose(lap_jet.func(x), expected, rtol=1e-6, atol=1e-6)

    coords = jnp.linspace(0.0, 1.0, 5)
    assert jnp.allclose(lap_jet.func((coords,)), 6.0 * coords, rtol=1e-6, atol=1e-6)


def test_div_diag_k_grad_backend_jet_constant_k_matches_laplacian():
    geom = Interval1d(0.0, 1.0)

    @geom.Function("x")
    def u(x):
        return x[0] ** 3

    @geom.Function("x")
    def k_vec(x):
        return jnp.array([1.0], dtype=float)

    op = div_diag_k_grad(u, k_vec, var="x", backend="jet")
    lap = laplacian(u, var="x", backend="jet")

    x = jnp.asarray([0.7])
    assert jnp.allclose(op.func(x), lap.func(x), rtol=1e-6, atol=1e-6)

    coords = jnp.linspace(0.0, 1.0, 6)
    expected = 6.0 * coords
    assert jnp.allclose(op.func((coords,)), expected, rtol=1e-6, atol=1e-6)


def test_bilaplacian_backend_jet_matches_ad_1d():
    geom = Interval1d(0.0, 1.0)

    @geom.Function("x")
    def u(x):
        return x[0] ** 4

    bilap_ad = bilaplacian(u, var="x", backend="ad")
    bilap_jet = bilaplacian(u, var="x", backend="jet")

    x = jnp.asarray([0.2])
    expected = 24.0
    assert jnp.allclose(bilap_ad.func(x), expected, rtol=1e-6, atol=1e-6)
    assert jnp.allclose(bilap_jet.func(x), expected, rtol=1e-6, atol=1e-6)

    coords = jnp.linspace(0.0, 1.0, 7)
    assert jnp.allclose(bilap_jet.func((coords,)), expected, rtol=1e-6, atol=1e-6)


def test_bilaplacian_backend_jet_matches_ad_2d_mixed_terms():
    geom = Square(center=(0.0, 0.0), side=2.0)

    @geom.Function("x")
    def u(x):
        return x[0] ** 4 + x[1] ** 4 + 2.0 * x[0] ** 2 * x[1] ** 2

    bilap_ad = bilaplacian(u, var="x", backend="ad")
    bilap_jet = bilaplacian(u, var="x", backend="jet")

    x = jnp.asarray([0.2, -0.1])
    expected = 64.0
    assert jnp.allclose(bilap_ad.func(x), expected, rtol=1e-6, atol=1e-6)
    assert jnp.allclose(bilap_jet.func(x), expected, rtol=1e-6, atol=1e-6)

    coords_x = jnp.linspace(-1.0, 1.0, 4)
    coords_y = jnp.linspace(-0.5, 0.5, 3)
    assert jnp.allclose(
        bilap_jet.func((coords_x, coords_y)), expected, rtol=1e-6, atol=1e-6
    )



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import coordax as cx
import jax.numpy as jnp

from phydrax._frozendict import frozendict
from phydrax.domain import Square, TimeInterval
from phydrax.operators.differential import div, grad, laplacian


def test_div_grad_equals_laplacian_scalar_point():
    geom = Square(center=(0.0, 0.0), side=2.0)

    @geom.Function("x")
    def f(x):
        return x[0] ** 2 + x[1] ** 2

    pts = frozendict({"x": cx.Field(jnp.array([2.0, 3.0]), dims=(None,))})
    out_divgrad = jnp.asarray(div(grad(f))(pts).data)
    out_lap = jnp.asarray(laplacian(f)(pts).data)
    assert jnp.allclose(out_divgrad, out_lap)


def test_div_grad_equals_laplacian_vector_point():
    geom = Square(center=(0.0, 0.0), side=2.0)

    @geom.Function("x")
    def f(x):
        return jnp.array([x[0] ** 2, x[1] ** 2])

    pts = frozendict({"x": cx.Field(jnp.array([2.0, 3.0]), dims=(None,))})
    out_divgrad = jnp.asarray(div(grad(f))(pts).data)
    out_lap = jnp.asarray(laplacian(f)(pts).data)
    assert jnp.allclose(out_divgrad, out_lap)


def test_div_grad_equals_laplacian_spacetime(sample_batch):
    geom = Square(center=(0.0, 0.0), side=2.0)
    dom = geom @ TimeInterval(0.0, 1.0)

    @dom.Function("x", "t")
    def f(x, t):
        return x[0] ** 2 + x[1] ** 2 + t

    divgrad = div(grad(f, var="x"), var="x")
    lap = laplacian(f, var="x")

    component = dom.component()
    batch = sample_batch(component, blocks=(("x",), ("t",)), num_points=(3, 4), key=0)
    out_divgrad = jnp.asarray(divgrad(batch).data)
    out_lap = jnp.asarray(lap(batch).data)
    assert jnp.allclose(out_divgrad, out_lap)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import numpy as np
import pytest

from phydrax.domain.geometry2d._primitives import (
    Circle,
    Ellipse,
    Polygon,
    Rectangle,
    Square,
    Triangle,
)


def test_circle():
    radius = 1.0
    circle = Circle(center=(0.0, 0.0), radius=radius)
    expected_area = np.pi * radius**2
    computed_area = float(circle.area)
    assert np.isclose(computed_area, expected_area, rtol=0.05)


def test_ellipse():
    x_radius = 2.0
    y_radius = 1.0
    ellipse = Ellipse(center=(0.0, 0.0), x_radius=x_radius, y_radius=y_radius)
    expected_area = np.pi * x_radius * y_radius
    computed_area = float(ellipse.area)
    assert np.isclose(computed_area, expected_area, rtol=0.05)


def test_rectangle():
    width = 3.0
    height = 2.0
    rectangle = Rectangle(center=(0.0, 0.0), width=width, height=height)
    expected_area = width * height
    computed_area = float(rectangle.area)
    assert np.isclose(computed_area, expected_area, rtol=0.05)


def test_square():
    side = 2.0
    square = Square(center=(0.0, 0.0), side=side)
    expected_area = side**2
    computed_area = float(square.area)
    assert np.isclose(computed_area, expected_area, rtol=0.05)


def test_polygon():
    vertices = [(0.0, 0.0), (2.0, 0.0), (1.0, 1.0)]
    polygon = Polygon(vertices=vertices)
    expected_area = 1.0  # Triangle area: 0.5 * base * height = 0.5 * 2 * 1
    computed_area = float(polygon.area)
    assert np.isclose(computed_area, expected_area, rtol=0.05)


def test_triangle():
    vertices = [(0.0, 0.0), (1.0, 0.0), (0.0, 1.0)]
    triangle = Triangle(vertices=vertices)
    expected_area = 0.5  # Right triangle area: 0.5 * base * height = 0.5 * 1 * 1
    computed_area = float(triangle.area)
    assert np.isclose(computed_area, expected_area, rtol=0.05)


def test_polygon_non_unique_vertices():
    vertices = [(0.0, 0.0), (1.0, 0.0), (1.0, 0.0), (0.0, 1.0)]
    with pytest.raises(ValueError, match="Non-unique vertices"):
        Polygon(vertices=vertices)


def test_polygon_self_intersection():
    vertices = [(0.0, 0.0), (1.0, 1.0), (1.0, 0.0), (0.0, 1.0)]
    with pytest.raises(ValueError, match="Self-intersection"):
        Polygon(vertices=vertices)


def test_triangle_invalid_vertices():
    vertices = [(0.0, 0.0), (1.0, 0.0)]  # Only 2 vertices
    with pytest.raises(ValueError, match="Triangle must have exactly 3 vertices"):
        Triangle(vertices=vertices)


def test_polygon_preserves_absolute_coordinates():
    # A non-centered polygon; ensure mesh bounds match input bounds
    vertices = [
        (1.0, 0.0),
        (1.0, 1.0),
        (0.8, 1.0),
        (0.8, 0.2),
        (0.6, 2.0),
        (0.6, 1.0),
        (0.4, 1.0),
        (0.4, 0.2),
        (0.2, 0.0),
    ]
    poly = Polygon(vertices=vertices)
    import numpy as np

    pts = np.asarray(poly.mesh.points)[:, :2]
    in_min = np.min(np.asarray(vertices), axis=0)
    in_max = np.max(np.asarray(vertices), axis=0)
    out_min = np.min(pts, axis=0)
    out_max = np.max(pts, axis=0)
    assert np.allclose(out_min, in_min, atol=1e-6)
    assert np.allclose(out_max, in_max, atol=1e-6)


def test_boundary_normals_circle():
    # Circle centered at origin with radius 1
    c = Circle(center=(0.0, 0.0), radius=1.0)
    # Points on boundary along axes
    from jax import numpy as jnp

    pts = jnp.array(
        [
            [1.0, 0.0],
            [-1.0, 0.0],
            [0.0, 1.0],
            [0.0, -1.0],
        ],
        dtype=float,
    )
    normals = c._boundary_normals(pts)
    expected = jnp.array(
        [
            [1.0, 0.0],
            [-1.0, 0.0],
            [0.0, 1.0],
            [0.0, -1.0],
        ]
    )
    import numpy as np

    assert np.allclose(normals, expected, atol=1e-3)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import coordax as cx
import jax.numpy as jnp
import jax.random as jr

from phydrax._frozendict import frozendict
from phydrax.domain import Interval1d, TimeInterval
from phydrax.operators.differential import caputo_time_fractional_dw


def test_caputo_time_fractional_dw_time_only_smoke():
    dom = TimeInterval(0.0, 2.0)

    @dom.Function("t")
    def u(t):
        return jnp.sin(t)

    D = caputo_time_fractional_dw(u, alpha=1.5)
    y = jnp.asarray(
        D(frozendict({"t": cx.Field(jnp.array(1.0), dims=())}), key=jr.key(0)).data
    )
    assert jnp.ndim(y) == 0
    assert jnp.isfinite(y)


def test_caputo_time_fractional_dw_broadcasts_over_space(sample_batch):
    dom = Interval1d(-1.0, 1.0) @ TimeInterval(0.0, 2.0)

    @dom.Function("t")
    def u(t):
        return jnp.cos(t)

    D = caputo_time_fractional_dw(u, alpha=1.25, M=64)
    component = dom.component()
    batch = sample_batch(component, blocks=(("x",), ("t",)), num_points=(2, 5), key=1)
    Y = jnp.asarray(D(batch, key=jr.key(1)).data)
    assert Y.shape == (2, 5)
    assert jnp.all(jnp.isfinite(Y))



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import jax.numpy as jnp

from phydrax.domain import (
    CosineAxisSpec,
    FourierAxisSpec,
    Interval1d,
    LegendreAxisSpec,
    SineAxisSpec,
)
from phydrax.operators.differential import laplacian, partial_n


def test_partial_n_basis_fourier_matches_closed_form_periodic():
    geom = Interval1d(0.0, 1.0)

    @geom.Function("x")
    def u(x):
        return jnp.sin(2.0 * jnp.pi * x[0])

    d2 = partial_n(
        u,
        var="x",
        order=2,
        backend="basis",
        basis="fourier",
        periodic=True,
    )
    coords = FourierAxisSpec(64).materialize(jnp.array(0.0), jnp.array(1.0)).nodes
    out = d2.func((coords,))
    expected = -((2.0 * jnp.pi) ** 2) * jnp.sin(2.0 * jnp.pi * coords)
    assert jnp.allclose(out, expected, rtol=1e-6, atol=1e-6)


def test_partial_n_fd_matches_closed_form_periodic_to_tolerance():
    geom = Interval1d(0.0, 1.0)

    @geom.Function("x")
    def u(x):
        return jnp.sin(2.0 * jnp.pi * x[0])

    d1 = partial_n(
        u,
        var="x",
        order=1,
        backend="fd",
        periodic=True,
    )
    coords = FourierAxisSpec(256).materialize(jnp.array(0.0), jnp.array(1.0)).nodes
    out = d1.func((coords,))
    expected = 2.0 * jnp.pi * jnp.cos(2.0 * jnp.pi * coords)
    assert jnp.allclose(out, expected, rtol=2e-3, atol=2e-3)


def test_partial_n_basis_poly_on_legendre_nodes_matches_closed_form():
    geom = Interval1d(-1.0, 1.0)

    @geom.Function("x")
    def u(x):
        return x[0] ** 3

    d2 = partial_n(
        u,
        var="x",
        order=2,
        backend="basis",
        basis="poly",
    )
    coords = LegendreAxisSpec(12).materialize(jnp.array(-1.0), jnp.array(1.0)).nodes
    out = d2.func((coords,))
    expected = 6.0 * coords
    assert jnp.allclose(out, expected, rtol=1e-6, atol=1e-6)


def test_laplacian_basis_matches_second_partial_in_1d():
    geom = Interval1d(0.0, 1.0)

    @geom.Function("x")
    def u(x):
        return jnp.sin(2.0 * jnp.pi * x[0])

    lap = laplacian(
        u,
        var="x",
        backend="basis",
        basis="fourier",
        periodic=True,
    )
    d2 = partial_n(
        u,
        var="x",
        order=2,
        backend="basis",
        basis="fourier",
        periodic=True,
    )
    coords = FourierAxisSpec(64).materialize(jnp.array(0.0), jnp.array(1.0)).nodes
    out_lap = lap.func((coords,))
    out_d2 = d2.func((coords,))
    assert jnp.allclose(out_lap, out_d2, rtol=1e-8, atol=1e-8)


def test_partial_n_basis_cosine_matches_closed_form():
    geom = Interval1d(0.0, 1.0)

    @geom.Function("x")
    def u(x):
        return jnp.cos(3.0 * jnp.pi * x[0])

    d1 = partial_n(
        u,
        var="x",
        order=1,
        backend="basis",
        basis="cosine",
    )
    coords = CosineAxisSpec(129).materialize(jnp.array(0.0), jnp.array(1.0)).nodes
    out = d1.func((coords,))
    expected = -3.0 * jnp.pi * jnp.sin(3.0 * jnp.pi * coords)
    assert jnp.allclose(out, expected, rtol=1e-6, atol=1e-6)


def test_partial_n_basis_sine_matches_closed_form():
    geom = Interval1d(0.0, 1.0)

    @geom.Function("x")
    def u(x):
        return jnp.sin(4.0 * jnp.pi * x[0])

    d1 = partial_n(
        u,
        var="x",
        order=1,
        backend="basis",
        basis="sine",
    )
    coords = SineAxisSpec(128).materialize(jnp.array(0.0), jnp.array(1.0)).nodes
    out = d1.func((coords,))
    expected = 4.0 * jnp.pi * jnp.cos(4.0 * jnp.pi * coords)
    assert jnp.allclose(out, expected, rtol=1e-6, atol=1e-6)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import jax.numpy as jnp
import jax.random as jr
import pytest

from phydrax.nn.models import SeparableFeynmaNN


def test_separable_feynmann_vector_input_shape():
    model = SeparableFeynmaNN(
        in_size=2,
        out_size=3,
        latent_size=4,
        width_size=8,
        depth=2,
        num_paths=2,
        key=jr.key(0),
    )
    x = jr.normal(jr.key(1), (2,))
    y = model(x)
    assert y.shape == (3,)
    assert jnp.all(jnp.isfinite(jnp.asarray(y)))


def test_separable_feynmann_coord_separable_shape():
    model = SeparableFeynmaNN(
        in_size=2,
        out_size="scalar",
        latent_size=3,
        width_size=8,
        depth=2,
        num_paths=2,
        key=jr.key(2),
    )
    x0 = jnp.linspace(0.0, 1.0, 5)
    x1 = jnp.linspace(-1.0, 1.0, 6)
    y = model((x0, x1))
    assert y.shape == (5, 6)
    assert jnp.all(jnp.isfinite(jnp.asarray(y)))


def test_separable_feynmann_scalar_requires_split_input():
    with pytest.raises(ValueError, match="requires in_size >= 2"):
        _ = SeparableFeynmaNN(
            in_size="scalar",
            out_size="scalar",
            width_size=8,
            depth=1,
            num_paths=2,
        )


def test_separable_feynmann_scalar_with_split_input():
    model = SeparableFeynmaNN(
        in_size="scalar",
        out_size="scalar",
        latent_size=2,
        split_input=2,
        width_size=6,
        depth=1,
        num_paths=2,
        key=jr.key(3),
    )
    x = jnp.asarray(0.25)
    y = model(x)
    assert y.shape == ()
    assert jnp.isfinite(y)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

from __future__ import annotations

from typing import Any, Literal

import jax.numpy as jnp
from jaxtyping import Array, Key

from ...._callable import _ensure_special_kwonly_args
from ...._doc import DOC_KEY0
from ..._utils import _canonical_size, _get_size, _get_value_shape, SizeLike
from ..core._base import _AbstractBaseModel, _AbstractStructuredInputModel


_Layout = Literal["value", "passthrough"]


def _flatten_value(x: Array, /, *, in_size: int | tuple[int, ...] | Literal["scalar"]):
    x_arr = jnp.asarray(x)
    if in_size == "scalar":
        if x_arr.shape == ():
            return x_arr, ()
        if x_arr.shape == (1,):
            return x_arr.reshape(()), ()
        raise ValueError(f"`x` must have scalar shape () (or (1,)), got {x_arr.shape}.")

    in_shape = _get_value_shape(in_size)
    if x_arr.shape != in_shape:
        raise ValueError(f"`x` must have shape {in_shape}, got {x_arr.shape}.")
    x_flat = x_arr.reshape((_get_size(in_size),))
    return x_flat, ()


def _reshape_value(
    y_flat: Array,
    /,
    *,
    leading_shape: tuple[int, ...],
    out_size: int | tuple[int, ...] | Literal["scalar"],
) -> Array:
    y_arr = jnp.asarray(y_flat)
    if out_size == "scalar":
        if leading_shape:
            raise ValueError(
                "Expected scalar output with no leading axes; got "
                f"leading_shape={leading_shape}."
            )
        if y_arr.shape == ():
            return y_arr
        if y_arr.shape == (1,):
            return jnp.squeeze(y_arr, axis=0)
        raise ValueError(
            "Wrapped module returned the wrong shape for scalar out_size. Expected "
            f"() or (1,), got {y_arr.shape}."
        )

    out_numel = _get_size(out_size)
    out_shape = _get_value_shape(out_size)
    if leading_shape:
        raise ValueError(
            "Expected unbatched output (no leading axes) for layout='value'. "
            f"Got leading_shape={leading_shape}."
        )
    if y_arr.shape != (out_numel,):
        raise ValueError(
            "Wrapped module returned the wrong shape. Expected "
            f"({out_numel},) (to reshape to {out_shape}), got {y_arr.shape}."
        )
    return y_arr.reshape(out_shape)


class EquinoxModel(_AbstractBaseModel):
    """Adapter for arbitrary Equinox/JAX callables with Phydrax model metadata.

    Default `layout="value"` treats `in_size/out_size` as **value shapes** and performs:
    (flatten value axes) -> (call wrapped module) -> (reshape back to value axes).

    Use `layout="passthrough"` to forward inputs/outputs unchanged.
    """

    module: Any
    in_size: int | tuple[int, ...] | Literal["scalar"]
    out_size: int | tuple[int, ...] | Literal["scalar"]
    layout: _Layout

    def __init__(
        self,
        module: Any,
        /,
        *,
        in_size: SizeLike,
        out_size: SizeLike,
        layout: _Layout = "value",
    ):
        self.module = _ensure_special_kwonly_args(module)
        self.in_size = _canonical_size(in_size)
        self.out_size = _canonical_size(out_size)
        self.layout = layout

    def __call__(
        self,
        x: Array,
        /,
        *,
        key: Key[Array, ""] = DOC_KEY0,
        iter_: Array | None = None,
        **kwargs: Any,
    ) -> Array:
        del iter_
        if self.layout == "passthrough":
            return self.module(x, key=key, **kwargs)

        x_flat, leading_shape = _flatten_value(x, in_size=self.in_size)
        y_flat = self.module(x_flat, key=key, **kwargs)
        return _reshape_value(y_flat, leading_shape=leading_shape, out_size=self.out_size)


class EquinoxStructuredModel(_AbstractStructuredInputModel):
    """Equinox/JAX callable adapter that supports structured (tuple) inputs."""

    module: Any
    in_size: int | tuple[int, ...] | Literal["scalar"]
    out_size: int | tuple[int, ...] | Literal["scalar"]
    layout: _Layout

    def __init__(
        self,
        module: Any,
        /,
        *,
        in_size: SizeLike,
        out_size: SizeLike,
        layout: _Layout = "passthrough",
    ):
        self.module = _ensure_special_kwonly_args(module)
        self.in_size = _canonical_size(in_size)
        self.out_size = _canonical_size(out_size)
        self.layout = layout

    def __call__(
        self,
        x: Array | tuple[Array, ...],
        /,
        *,
        key: Key[Array, ""] = DOC_KEY0,
        iter_: Array | None = None,
        **kwargs: Any,
    ) -> Array:
        del iter_
        if self.layout == "passthrough":
            if isinstance(x, tuple):
                x_in = tuple(jnp.asarray(a) for a in x)
            else:
                x_in = jnp.asarray(x)
            return self.module(x_in, key=key, **kwargs)

        if isinstance(x, tuple):
            parts = []
            for a in x:
                arr = jnp.asarray(a)
                if arr.ndim > 1:
                    raise ValueError(
                        "EquinoxStructuredModel(layout='value') expects tuple inputs to be "
                        "scalars/vectors (ndim<=1). Got an input with shape "
                        f"{arr.shape}."
                    )
                parts.append(arr.reshape((-1,)))
            x_arr = parts[0] if len(parts) == 1 else jnp.concatenate(parts, axis=0)
        else:
            x_arr = jnp.asarray(x)

        x_flat, leading_shape = _flatten_value(x_arr, in_size=self.in_size)
        y_flat = self.module(x_flat, key=key, **kwargs)
        return _reshape_value(y_flat, leading_shape=leading_shape, out_size=self.out_size)


__all__ = ["EquinoxModel", "EquinoxStructuredModel"]



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import functools as ft
from abc import abstractmethod
from collections.abc import Callable, Sequence
from typing import Literal

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Bool, Float, Key

from .._doc import DOC_KEY0
from .._strict import AbstractAttribute
from ._domain import _AbstractUnaryDomain


class _AbstractGeometry(_AbstractUnaryDomain):
    """Abstract (spatial) geometry."""

    adf: AbstractAttribute[Callable[[Array], Array]]

    @property
    @abstractmethod
    def volume(self) -> Array:
        raise NotImplementedError

    @property
    @abstractmethod
    def spatial_dim(self) -> int:
        raise NotImplementedError

    @property
    def dim(self) -> int:
        return self.spatial_dim

    @property
    def label(self) -> str:
        return "x"

    @property
    def var_dim(self) -> int:
        return int(self.spatial_dim)

    @property
    @abstractmethod
    def bounds(self) -> Float[Array, "2 spatial_dim"]:
        raise NotImplementedError

    @property
    def time(self) -> Literal[False]:
        return False

    @ft.cached_property
    def mesh_bounds(self) -> Float[Array, "2 spatial_dim"]:
        """Axis-aligned bounding box as `[[mins...], [maxs...]]` (raw values)."""
        bounds = jnp.asarray(self.bounds, dtype=float)
        sd = int(self.spatial_dim)
        if bounds.shape != (2, sd):
            raise ValueError(
                f"{type(self).__name__}.bounds must have shape (2, {sd}), got {bounds.shape}."
            )
        return bounds

    @ft.cached_property
    def volume_proportion(self) -> Float[Array, ""]:
        """Fraction of the AABB volume occupied by the geometry (defaults to 1.0)."""
        return jnp.array(1.0, dtype=float)

    @property
    def boundary_measure_value(self) -> Array:
        """Total boundary measure value (boundary length / surface area).

        Concrete geometries should override this where applicable. For 1D geometries,
        this defaults to counting measure on the two endpoints (value = 2).
        """
        if self.spatial_dim == 1:
            return jnp.array(2.0, dtype=float)
        raise NotImplementedError(
            f"{type(self).__name__} must implement `boundary_measure_value`."
        )

    @abstractmethod
    def estimate_boundary_subset_measure(
        self,
        where: Callable[[Array], Bool[Array, ""]],
        *,
        num_samples: int = 4096,
        key: Key[Array, ""] = DOC_KEY0,
    ) -> Array:
        """Estimate boundary subset measure of {x: where(x)=True}."""
        raise NotImplementedError

    def _check_points_on_boundary(self, points: Array) -> Array:
        return eqx.error_if(
            points,
            pred=~self._on_boundary(points),
            msg="All points must be on the boundary of the domain.",
        )

    @abstractmethod
    def sample_interior(
        self,
        num_points: int,
        *,
        where: Callable | None = None,
        sampler: str = "latin_hypercube",
        key: Key[Array, ""] = DOC_KEY0,
    ) -> Array:
        raise NotImplementedError

    @abstractmethod
    def sample_boundary(
        self,
        num_points: int,
        *,
        where: Callable | None = None,
        sampler: str = "latin_hypercube",
        key: Key[Array, ""] = DOC_KEY0,
    ) -> Array:
        raise NotImplementedError

    @abstractmethod
    def _sample_interior_separable(
        self,
        num_points: int | Sequence[int],
        *,
        sampler: str = "latin_hypercube",
        where: Callable | None = None,
        key: Key[Array, ""] = DOC_KEY0,
    ) -> tuple[tuple[Array, ...], Bool[Array, "..."]]:
        """Internal helper for separable interior sampling."""
        raise NotImplementedError

    @abstractmethod
    def _contains(self, points: Array) -> Bool[Array, " num_points"]:
        raise NotImplementedError

    @abstractmethod
    def _on_boundary(self, points: Array) -> Bool[Array, " num_points"]:
        raise NotImplementedError

    @abstractmethod
    def _boundary_normals(self, points: Array) -> Float[Array, "num_points spatial_dim"]:
        raise NotImplementedError



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

from __future__ import annotations

from typing import Any

import coordax as cx
import jax.numpy as jnp
from jaxtyping import Array, Key

from ..._doc import DOC_KEY0
from ...domain._components import DomainComponent, DomainComponentUnion
from ...domain._function import DomainFunction
from ...domain._structure import CoordSeparableBatch, PointsBatch, QuadratureBatch
from ..integral._batch_ops import integral, mean


def spatial_mean(
    u: DomainFunction,
    batch: PointsBatch | CoordSeparableBatch | tuple[PointsBatch, ...],
    /,
    *,
    component: DomainComponent | DomainComponentUnion,
    quadrature: QuadratureBatch | tuple[QuadratureBatch | None, ...] | None = None,
    over: str | tuple[str, ...] | None = "x",
    key: Key[Array, ""] = DOC_KEY0,
    **kwargs: Any,
) -> cx.Field:
    r"""Estimate the spatial mean of a field.

    Computes the Monte Carlo / quadrature estimate of

    $$
    \langle u \rangle = \frac{1}{\mu(\Omega)}\int_{\Omega} u(x)\,d\mu(x),
    $$

    where $\Omega$ is the subset described by `component` and the integration axes
    are selected by `over` (default `"x"`).

    **Arguments:**

    - `u`: Integrand `DomainFunction`.
    - `batch`: `PointsBatch`/`CoordSeparableBatch` used for sampling (or a tuple for unions).
    - `component`: `DomainComponent` or `DomainComponentUnion` defining the integration region.
    - `quadrature`: Optional `QuadratureBatch` for paired sampling.
    - `over`: Which axes/labels to reduce over (defaults to `"x"`).
    - `key`: PRNG key forwarded to callables.
    - `kwargs`: Extra keyword arguments forwarded to `u` and component callables.

    **Returns:**

    - A `coordax.Field` containing the mean value (with remaining named axes, if any).
    """
    return mean(
        u, batch, component=component, quadrature=quadrature, over=over, key=key, **kwargs
    )


def spatial_inner_product(
    u: DomainFunction,
    v: DomainFunction,
    batch: PointsBatch | CoordSeparableBatch | tuple[PointsBatch, ...],
    /,
    *,
    component: DomainComponent | DomainComponentUnion,
    quadrature: QuadratureBatch | tuple[QuadratureBatch | None, ...] | None = None,
    over: str | tuple[str, ...] | None = "x",
    key: Key[Array, ""] = DOC_KEY0,
    **kwargs: Any,
) -> cx.Field:
    r"""Estimate an $L^2$-type inner product.

    Forms the pointwise Euclidean/Frobenius product and integrates:

    $$
    \langle u, v \rangle = \int_{\Omega} u(x)\cdot v(x)\,d\mu(x),
    $$

    where `u(x)·v(x)` is implemented as `sum(u * v)` over the value dimensions.
    Domains are joined before evaluation.

    **Arguments:**

    - `u`, `v`: Input `DomainFunction`s.
    - `batch`, `component`, `quadrature`, `over`, `key`, `kwargs`: As in `spatial_mean`.

    **Returns:**

    - A `coordax.Field` containing the scalar inner product value.
    """
    joined = u.domain.join(v.domain)
    u2 = u.promote(joined)
    v2 = v.promote(joined)

    deps = tuple(lbl for lbl in joined.labels if (lbl in u2.deps) or (lbl in v2.deps))
    idx = {lbl: i for i, lbl in enumerate(deps)}
    u_pos = tuple(idx[lbl] for lbl in u2.deps)
    v_pos = tuple(idx[lbl] for lbl in v2.deps)

    def _ip(*args, key=None, **kw):
        u_args = [args[i] for i in u_pos]
        v_args = [args[i] for i in v_pos]
        uu = jnp.asarray(u2.func(*u_args, key=key, **kw))
        vv = jnp.asarray(v2.func(*v_args, key=key, **kw))
        return jnp.sum(uu * vv)

    ip = DomainFunction(domain=joined, deps=deps, func=_ip, metadata={})
    return integral(
        ip,
        batch,
        component=component,
        quadrature=quadrature,
        over=over,
        key=key,
        **kwargs,
    )


def spatial_lp_norm(
    u: DomainFunction,
    batch: PointsBatch | CoordSeparableBatch | tuple[PointsBatch, ...],
    /,
    *,
    p: float = 2.0,
    component: DomainComponent | DomainComponentUnion,
    quadrature: QuadratureBatch | tuple[QuadratureBatch | None, ...] | None = None,
    over: str | tuple[str, ...] | None = "x",
    key: Key[Array, ""] = DOC_KEY0,
    **kwargs: Any,
) -> cx.Field:
    r"""Estimate an $L^p$ norm over space.

    Computes

    $$
    \|u\|_{L^p(\Omega)} = \left(\int_{\Omega} \|u(x)\|_2^p\,d\mu(x)\right)^{1/p},
    $$

    where $\|u(x)\|_2$ is the Euclidean norm of the (possibly vector/tensor) value at
    $x$, flattened.

    **Arguments:**

    - `u`: Input `DomainFunction`.
    - `batch`, `component`, `quadrature`, `over`, `key`, `kwargs`: As in `spatial_mean`.
    - `p`: Norm exponent $p>0$.

    **Returns:**

    - A `coordax.Field` containing the scalar $L^p$ norm value.
    """
    if p <= 0:
        raise ValueError("p must be positive.")
    p_ = float(p)

    def _pow_norm(*args, key=None, **kw):
        val = jnp.asarray(u.func(*args, key=key, **kw))
        flat = val.reshape((-1,))
        return jnp.power(jnp.linalg.norm(flat), p_)

    integrand = DomainFunction(domain=u.domain, deps=u.deps, func=_pow_norm, metadata={})
    val = integral(
        integrand,
        batch,
        component=component,
        quadrature=quadrature,
        over=over,
        key=key,
        **kwargs,
    )
    return cx.Field(jnp.power(jnp.asarray(val.data), 1.0 / p_), dims=val.dims)


def spatial_l2_norm(
    u: DomainFunction,
    batch: PointsBatch | CoordSeparableBatch | tuple[PointsBatch, ...],
    /,
    *,
    component: DomainComponent | DomainComponentUnion,
    quadrature: QuadratureBatch | tuple[QuadratureBatch | None, ...] | None = None,
    over: str | tuple[str, ...] | None = "x",
    key: Key[Array, ""] = DOC_KEY0,
    **kwargs: Any,
) -> cx.Field:
    r"""Estimate the $L^2$ norm over space.

    Equivalent to `spatial_lp_norm(..., p=2)`.
    """
    return spatial_lp_norm(
        u,
        batch,
        p=2.0,
        component=component,
        quadrature=quadrature,
        over=over,
        key=key,
        **kwargs,
    )


__all__ = [
    "spatial_inner_product",
    "spatial_l2_norm",
    "spatial_lp_norm",
    "spatial_mean",
]



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import coordax as cx
import jax.numpy as jnp

from phydrax._frozendict import frozendict
from phydrax.domain import Square
from phydrax.operators.differential import directional_derivative, div, grad, laplacian


def test_grad_complex_scalar_point():
    geom = Square(center=(0.0, 0.0), side=2.0)

    @geom.Function("x")
    def u(x):
        return x[0] ** 2 + 1j * x[1] ** 2

    g = grad(u)
    pts = frozendict({"x": cx.Field(jnp.array([2.0, 3.0]), dims=(None,))})
    out = jnp.asarray(g(pts).data)
    assert jnp.allclose(out, jnp.array([4.0, 6.0j]))


def test_directional_derivative_complex_scalar_point():
    geom = Square(center=(0.0, 0.0), side=2.0)

    @geom.Function("x")
    def u(x):
        return x[0] ** 2 + 1j * x[1] ** 2

    v = geom.Function()(jnp.array([1.0, 1.0]))
    dd = directional_derivative(u, v)

    pts = frozendict({"x": cx.Field(jnp.array([2.0, 3.0]), dims=(None,))})
    out = jnp.asarray(dd(pts).data)
    assert jnp.allclose(out, 4.0 + 6.0j)


def test_div_complex_vector_field_point():
    geom = Square(center=(0.0, 0.0), side=2.0)

    @geom.Function("x")
    def u(x):
        return jnp.array([x[0], 1j * x[1]])

    div_u = div(u)
    pts = frozendict({"x": cx.Field(jnp.array([0.3, -1.1]), dims=(None,))})
    out = jnp.asarray(div_u(pts).data)
    assert jnp.allclose(out, 1.0 + 1.0j)


def test_laplacian_complex_scalar_point():
    geom = Square(center=(0.0, 0.0), side=2.0)

    @geom.Function("x")
    def u(x):
        return x[0] ** 2 + 1j * x[1] ** 2

    lap_u = laplacian(u)
    pts = frozendict({"x": cx.Field(jnp.array([0.3, -1.1]), dims=(None,))})
    out = jnp.asarray(lap_u(pts).data)
    assert jnp.allclose(out, 2.0 + 2.0j)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import jax.numpy as jnp
import jax.random as jr

from phydrax.constraints import PointSetConstraint
from phydrax.domain import Interval1d


def test_pointset_penalty_mean_and_sum():
    geom = Interval1d(0.0, 1.0)
    component = geom.component()

    points = {"x": jnp.array([[0.0], [0.5], [1.0]], dtype=float)}
    u = geom.Function()(0.0)

    def residual(functions):
        return functions["u"] - 1.0

    c_mean = PointSetConstraint.from_points(
        component=component,
        points=points,
        residual=residual,
        reduction="mean",
        weight=2.0,
    )
    loss_mean = c_mean.loss({"u": u}, key=jr.key(0))
    assert jnp.allclose(loss_mean, 2.0)

    c_sum = PointSetConstraint.from_points(
        component=component,
        points=points,
        residual=residual,
        reduction="sum",
        weight=2.0,
    )
    loss_sum = c_sum.loss({"u": u}, key=jr.key(0))
    assert jnp.allclose(loss_sum, 6.0)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import coordax as cx
import jax.numpy as jnp

from phydrax._frozendict import frozendict
from phydrax.domain import DomainFunction, Square, TimeInterval
from phydrax.operators.differential import div


def test_div_vector_field_point():
    geom = Square(center=(0.0, 0.0), side=2.0)

    @geom.Function("x")
    def u(x):
        return jnp.array([x[0], x[1]])

    d = div(u)
    pts = frozendict({"x": cx.Field(jnp.array([2.0, 3.0]), dims=(None,))})
    out = jnp.asarray(d(pts).data)
    assert jnp.allclose(out, jnp.array(2.0))


def test_div_spacetime_var_x_ignores_t(sample_batch):
    dom = Square(center=(0.0, 0.0), side=2.0) @ TimeInterval(0.0, 1.0)

    @dom.Function("x")
    def u(x):
        return jnp.array([x[0], x[1]])

    d = div(u, var="x")
    component = dom.component()
    batch = sample_batch(component, blocks=(("x",), ("t",)), num_points=(4, 3), key=0)
    out = jnp.asarray(d(batch).data)
    assert out.shape == (4, 3)
    assert jnp.allclose(out, 2.0)


def test_div_coord_separable_constant(sample_coord_separable):
    geom = Square(center=(0.0, 0.0), side=2.0)
    component = geom.component()
    batch = sample_coord_separable(component, {"x": (7, 6)}, dense_blocks=(), key=1)

    @geom.Function("x")
    def u(x):
        x, y = x
        return jnp.stack([2.0 * x, 3.0 * y], axis=-1)

    d = div(u)
    out = jnp.asarray(d(batch).data)
    assert jnp.allclose(out, 5.0, atol=1e-6)


def test_div_preserves_metadata():
    geom = Square(center=(0.0, 0.0), side=2.0)
    u = DomainFunction(
        domain=geom,
        deps=("x",),
        func=lambda x: jnp.array([x[0], x[1]]),
        metadata={"tag": 1},
    )
    out = div(u)
    assert out.metadata == u.metadata



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import jax.numpy as jnp
import jax.random as jr
import numpy as np
import pytest

from phydrax.domain.geometry1d._primitives import Interval1d


def test_interval():
    start, end = 0.0, 2.0
    interval = Interval1d(start=start, end=end)
    assert np.isclose(float(interval.length), end - start, atol=1e-6)


def test_line_segment_invalid_endpoints():
    start, end = 2.0, 0.0  # Invalid since start >= end
    with pytest.raises(ValueError, match="`start` must be less than `end`."):
        Interval1d(start=start, end=end)


@pytest.fixture
def simple_interval():
    return Interval1d(start=0.0, end=1.0)


def test_sample_interior(simple_interval):
    num_points = 100
    sampled_points = simple_interval.sample_interior(num_points=num_points)
    assert sampled_points.shape == (num_points, 1)
    assert np.all(sampled_points >= simple_interval.start)
    assert np.all(sampled_points <= simple_interval.end)


def test_sample_boundary(simple_interval):
    num_points = 50
    key = jr.key(1)
    sampled_points = simple_interval.sample_boundary(num_points=num_points, key=key)
    assert sampled_points.shape == (num_points, 1)
    boundary_values = [float(simple_interval.start), float(simple_interval.end)]
    assert np.all(np.isin(np.asarray(sampled_points).flatten(), boundary_values))


def test_contains_method(simple_interval):
    inside_point = jnp.array([[0.5]])
    outside_point = jnp.array([[1.5]])
    assert simple_interval._contains(inside_point)[0]
    assert not simple_interval._contains(outside_point)[0]


def test_on_boundary_method(simple_interval):
    boundary_point = jnp.array([[1.0]])
    interior_point = jnp.array([[0.5]])
    assert simple_interval._on_boundary(boundary_point)[0]
    assert not simple_interval._on_boundary(interior_point)[0]


def test_boundary_normals(simple_interval):
    points = jnp.array([[0.0], [1.0]])
    expected_normals = np.array([[-1.0], [1.0]])
    computed_normals = simple_interval._boundary_normals(points)
    assert np.allclose(computed_normals, expected_normals, atol=1e-6)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

from typing import cast

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr

from phydrax.constraints import (
    ContinuousInitialConstraint,
    ContinuousODEConstraint,
    DiscreteODEConstraint,
    FunctionalConstraint,
    IntegralEqualityConstraint,
    PointSetConstraint,
)
from phydrax.constraints._continuous_interior import ContinuousPointwiseInteriorConstraint
from phydrax.constraints._discrete_interior import DiscreteInteriorDataConstraint
from phydrax.constraints._functional_initial import DiscreteInitialConstraint
from phydrax.domain import (
    DomainComponentUnion,
    FixedStart,
    Interval1d,
    ProductStructure,
    TimeInterval,
)
from phydrax.operators.differential import div_diag_k_grad, dt, laplacian


def _jit_loss(constraint, functions):
    loss_fn = eqx.filter_jit(lambda k: constraint.loss(functions, key=k))
    return loss_fn(jr.key(0))


def test_continuous_initial_coord_separable_spatial():
    geom = Interval1d(0.0, 1.0)
    time = TimeInterval(0.0, 1.0)
    domain = geom @ time
    component = domain.component({"t": FixedStart()})
    structure = ProductStructure((("x",),))

    @domain.Function("x", "t")
    def u(x, t):
        return 0.0

    constraint = ContinuousInitialConstraint(
        "u",
        component,
        func=0.0,
        num_points=(),
        structure=structure,
        coord_separable={"x": 4},
    )
    assert _jit_loss(constraint, {"u": u}) < 1e-6


def test_integral_constraint_coord_separable_constant():
    geom = Interval1d(0.0, 1.0)
    time = TimeInterval(0.0, 1.0)
    domain = geom @ time
    component = domain.component()
    structure = ProductStructure((("x", "t"),))

    @domain.Function("x", "t")
    def u(x, t):
        return 1.0

    constraint = IntegralEqualityConstraint.from_operator(
        component=component,
        operator=lambda f: f,
        constraint_vars="u",
        num_points=6,
        structure=structure,
        coord_separable={"x": 5},
        equal_to=1.0,
    )
    assert _jit_loss(constraint, {"u": u}) < 1e-6


def test_integral_constraint_over_axis_constant():
    geom = Interval1d(0.0, 1.0)
    time = TimeInterval(0.0, 1.0)
    domain = geom @ time
    component = domain.component()
    structure = ProductStructure((("x",), ("t",)))
    num_x = 8
    num_t = 6

    @domain.Function("x", "t")
    def u(x, t):
        return 1.0

    expected = 1.0 / float(num_t)
    constraint = IntegralEqualityConstraint.from_operator(
        component=component,
        operator=lambda f: f,
        constraint_vars="u",
        num_points=(num_x, num_t),
        structure=structure,
        over="x",
        equal_to=expected,
    )
    assert _jit_loss(constraint, {"u": u}) < 1e-6


def test_functional_constraint_union_zero_loss():
    geom = Interval1d(0.0, 1.0)
    c1 = geom.component(where={"x": lambda p: p[0] < 0.5})
    c2 = geom.component(where={"x": lambda p: p[0] >= 0.5})
    union = DomainComponentUnion((c1, c2))
    structure = ProductStructure((("x",),))

    @geom.Function("x")
    def u(x):
        return 0.0

    constraint = FunctionalConstraint.from_operator(
        component=union,
        operator=lambda f: f,
        constraint_vars="u",
        num_points=8,
        structure=structure,
    )
    assert _jit_loss(constraint, {"u": u}) < 1e-6


def test_where_all_masks_interior_constraint():
    geom = Interval1d(0.0, 1.0)
    structure = ProductStructure((("x",),))

    @geom.Function("x")
    def u(x):
        return jnp.where(x[..., 0] < 0.5, 0.0, 1.0)

    @geom.Function("x")
    def mask(x):
        return jnp.where(x[..., 0] < 0.5, 1.0, 0.0)

    constraint = ContinuousPointwiseInteriorConstraint(
        "u",
        geom,
        operator=lambda f: f,
        num_points=16,
        structure=structure,
        where_all=mask,
    )
    assert _jit_loss(constraint, {"u": u}) < 1e-6


def test_discrete_interior_sensor_track_custom_weights_zero():
    geom = Interval1d(0.0, 1.0)
    time = TimeInterval(0.0, 1.0)
    domain = geom @ time
    structure = ProductStructure((("x", "t"),))

    @domain.Function("x", "t")
    def u(x, t):
        return 1.0

    sensors = jnp.array([[0.2], [0.8]], dtype=float)
    times = jnp.array([0.25, 0.5, 0.75], dtype=float)
    sensor_values = jnp.ones((2, 3), dtype=float)

    constraint = DiscreteInteriorDataConstraint(
        "u",
        domain,
        sensors=sensors,
        times=times,
        sensor_values=sensor_values,
        num_points=12,
        structure=structure,
        idw_exponent=3.0,
        eps_snap=1e-6,
        lengthscales={"x": 0.3},
    )
    assert _jit_loss(constraint, {"u": u}) < 1e-6


def test_pointset_constraint_weighted_sum():
    geom = Interval1d(0.0, 1.0)
    component = geom.component()

    @geom.Function("x")
    def u(x):
        return 1.0

    points = {"x": jnp.array([[0.1], [0.4], [0.9]], dtype=float)}

    constraint = PointSetConstraint.from_points(
        component=component,
        points=points,
        residual=lambda fns: fns["u"],
        weight=2.0,
        reduction="sum",
    )
    loss = _jit_loss(constraint, {"u": u})
    assert jnp.allclose(loss, 6.0)


def test_discrete_initial_constraint_forward_mode():
    time = TimeInterval(0.0, 1.0).relabel("tau")
    component = time.component({"tau": FixedStart()})

    @time.Function("tau")
    def u(tau):
        return tau**2

    constraint = DiscreteInitialConstraint(
        "u",
        component,
        evolution_var="tau",
        points={},
        values=0.0,
        time_derivative_order=1,
        mode="forward",
    )
    assert _jit_loss(constraint, {"u": u}) < 1e-6


def test_ode_constraints_relabel_nonuniform_times():
    time = TimeInterval(0.0, 1.0).relabel("tau")
    structure = ProductStructure((("tau",),))

    @time.Function("tau")
    def u(tau):
        return tau**2

    @time.Function("tau")
    def target(tau):
        return 2.0 * tau

    def operator(f):
        return dt(f, var="tau") - target

    continuous = ContinuousODEConstraint(
        "u",
        cast(TimeInterval, time),
        operator,
        num_points=32,
        structure=structure,
    )
    assert _jit_loss(continuous, {"u": u}) < 1e-6

    times = jnp.array([0.0, 0.1, 0.4, 1.0], dtype=float)
    discrete = DiscreteODEConstraint(
        "u",
        cast(TimeInterval, time),
        operator,
        times=times,
    )
    assert _jit_loss(discrete, {"u": u}) < 1e-6


def test_integral_constraint_union_zero_loss():
    geom = Interval1d(0.0, 1.0)
    left = geom.component(where={"x": lambda p: p[0] < 0.5})
    right = geom.component(where={"x": lambda p: p[0] >= 0.5})
    union = DomainComponentUnion((left, right))
    structure = ProductStructure((("x",),))

    @geom.Function("x")
    def u(x):
        return 0.0

    constraint = IntegralEqualityConstraint.from_operator(
        component=union,
        operator=lambda f: f,
        constraint_vars="u",
        num_points=8,
        structure=structure,
        equal_to=0.0,
    )
    assert _jit_loss(constraint, {"u": u}) < 1e-6


def test_integral_constraint_where_zero_mask():
    geom = Interval1d(0.0, 1.0)
    component = geom.component(where={"x": lambda p: p * 0.0})
    structure = ProductStructure((("x",),))

    @geom.Function("x")
    def u(x):
        return 1.0

    constraint = IntegralEqualityConstraint.from_operator(
        component=component,
        operator=lambda f: f,
        constraint_vars="u",
        num_points=16,
        structure=structure,
        equal_to=0.0,
    )
    assert _jit_loss(constraint, {"u": u}) < 1e-6


def test_discrete_interior_sensor_track_coord_separable_multilabel():
    x_dom = Interval1d(0.0, 1.0)
    y_dom = Interval1d(0.0, 1.0).relabel("y")
    time = TimeInterval(0.0, 1.0)
    domain = x_dom @ y_dom @ time
    structure = ProductStructure((("x", "y", "t"),))

    @domain.Function("x", "y", "t")
    def u(x, y, t):
        return 1.0

    sensors = {
        "x": jnp.array([[0.2], [0.8]], dtype=float),
        "y": jnp.array([[0.3], [0.7]], dtype=float),
    }
    times = jnp.array([0.25, 0.75], dtype=float)
    sensor_values = jnp.ones((2, 2), dtype=float)

    constraint = DiscreteInteriorDataConstraint(
        "u",
        domain,
        sensors=sensors,
        times=times,
        sensor_values=sensor_values,
        num_points=12,
        structure=structure,
        coord_separable={"x": 4},
    )
    assert _jit_loss(constraint, {"u": u}) < 1e-6


def test_coord_separable_laplacian_jet_zero():
    geom = Interval1d(0.0, 1.0)
    structure = ProductStructure((("x",),))

    @geom.Function("x")
    def u(x):
        return 1.0

    constraint = ContinuousPointwiseInteriorConstraint(
        "u",
        geom,
        operator=lambda f: laplacian(f, var="x", backend="jet"),
        num_points=16,
        structure=structure,
        coord_separable={"x": 8},
    )
    assert _jit_loss(constraint, {"u": u}) < 1e-6


def test_coord_separable_div_diag_k_grad_jet_zero():
    geom = Interval1d(0.0, 1.0)
    structure = ProductStructure((("x",),))

    @geom.Function("x")
    def u(x):
        return 1.0

    @geom.Function("x")
    def k_vec(x):
        return jnp.array([1.0], dtype=float)

    constraint = ContinuousPointwiseInteriorConstraint(
        "u",
        geom,
        operator=lambda f: div_diag_k_grad(f, k_vec, var="x", backend="jet"),
        num_points=16,
        structure=structure,
        coord_separable={"x": 8},
    )
    assert _jit_loss(constraint, {"u": u}) < 1e-6



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

"""
Spectral-Scaled Broyden (scalar inverse-Hessian) gradient transformation.

Implements a simple spectral step based on the alternate Barzilai–Borwein
formula often associated with Broyden-type updates:

  gamma_k = (s_k^T s_k) / (s_k^T y_k)

where s_k = x_k - x_{k-1}, y_k = g_k - g_{k-1}. The resulting update is

  u_k = - gamma_k * g_k

This keeps the Optax interface and can be chained with standard learning-rate
and weight-decay transformations.
"""

from typing import Any, NamedTuple

import jax
import jax.numpy as jnp
from optax._src import base, transform

from ._extra_args import GradientTransformationExtraArgs


def _line_search_strong_wolfe(
    value_fn, params, f0, g0, p, *, c1=1e-4, c2=0.9, alpha0=1.0, max_iters=20
):
    def tree_add(x, a, p):
        return jax.tree_util.tree_map(lambda xi, pi: xi + a * pi, x, p)

    def tree_dot(a, b):
        leaves = jax.tree_util.tree_map(lambda x, y: jnp.vdot(x, y), a, b)
        s = jax.tree_util.tree_reduce(lambda acc, v: acc + v, leaves, 0.0)
        return jnp.real(s) if jnp.iscomplexobj(s) else s

    g0p = tree_dot(g0, p)
    if g0p >= 0:
        p = jax.tree_util.tree_map(lambda x: -x, g0)
        g0p = tree_dot(g0, p)

    alpha = alpha0
    for _ in range(max_iters):
        x_new = tree_add(params, alpha, p)
        f_new = value_fn(x_new)
        if f_new > f0 + c1 * alpha * g0p:
            alpha *= 0.5
            continue
        g_new = jax.grad(value_fn)(x_new)
        gnp = tree_dot(g_new, p)
        if jnp.abs(gnp) <= c2 * jnp.abs(g0p):
            return alpha, p
        if gnp < 0:
            alpha *= 1.1
        else:
            alpha *= 0.5
    return alpha, p


def _tree_dot(a, b):
    def dot(x, y):
        x = jnp.asarray(x)
        y = jnp.asarray(y)
        v = jnp.vdot(x, y)
        return jnp.real(v) if jnp.iscomplexobj(v) else v

    leaves = jax.tree_util.tree_map(dot, a, b)
    return jax.tree_util.tree_reduce(lambda acc, v: acc + v, leaves, 0.0)


class SSBroydenState(NamedTuple):
    prev_params: base.Params
    prev_grad: base.Updates
    gamma: jnp.ndarray
    step: jnp.ndarray


def _as_ssbroyden_state(state: base.OptState, /) -> SSBroydenState:
    if not isinstance(state, SSBroydenState):
        raise TypeError("Expected SSBroydenState for optimizer state.")
    return state


def _split_decay_state(state: base.OptState, /) -> tuple[SSBroydenState, base.OptState]:
    if (
        not isinstance(state, tuple)
        or len(state) != 2
        or not isinstance(state[0], SSBroydenState)
    ):
        raise TypeError("Expected (SSBroydenState, decay_state) for optimizer state.")
    assert isinstance(state[1], base.OptState)
    return state[0], state[1]


def _ssbroyden_core(
    eps: float = 1e-12,
    gamma0: float = 1.0,
    gamma_min: float = 1e-6,
    gamma_max: float = 1e6,
    init_lr: float = 1.0,
    wolfe: bool = True,
    c1: float = 1e-4,
    c2: float = 0.9,
):
    def init_fn(params: base.Params) -> SSBroydenState:
        zeros_like = jax.tree_util.tree_map(lambda p: jnp.zeros_like(p), params)
        return SSBroydenState(
            prev_params=params,
            prev_grad=zeros_like,
            gamma=jnp.array(gamma0),
            step=jnp.array(0, dtype=jnp.int32),
        )

    def update_fn(
        updates: base.Updates,
        state: base.OptState,
        params: base.Params | None = None,
        **extra_args: Any,
    ) -> tuple[base.Updates, base.OptState]:
        grads = updates
        state_ss = _as_ssbroyden_state(state)
        params_ = params if params is not None else state_ss.prev_params
        step = state_ss.step + 1
        gamma = state_ss.gamma

        s = jax.tree_util.tree_map(lambda p, p0: p - p0, params_, state_ss.prev_params)
        y = jax.tree_util.tree_map(lambda g, g0: g - g0, grads, state_ss.prev_grad)

        sts = _tree_dot(s, s)
        sty = _tree_dot(s, y)

        # gamma = (s^T s) / (s^T y); protect against division by ~0 using cond
        def _upd(_):
            return jnp.clip(sts / (sty + jnp.sign(sty) * eps), gamma_min, gamma_max)

        gamma_new = jax.lax.cond(jnp.abs(sty) > eps, _upd, lambda _: gamma, operand=None)

        direction = jax.tree_util.tree_map(lambda g: -gamma_new * g, grads)
        if wolfe and ("value_fn" in extra_args) and ("value" in extra_args):
            alpha, direction = _line_search_strong_wolfe(
                extra_args["value_fn"],
                params_,
                extra_args["value"],
                grads,
                direction,
                c1=c1,
                c2=c2,
                alpha0=init_lr,
            )
            updates = jax.tree_util.tree_map(lambda d: alpha * d, direction)
        else:
            updates = jax.tree_util.tree_map(lambda d: init_lr * d, direction)
        new_state = SSBroydenState(
            prev_params=params_,
            prev_grad=grads,
            gamma=gamma_new,
            step=step,
        )
        return updates, new_state

    return GradientTransformationExtraArgs(init_fn, update_fn)


def ssbroyden(
    learning_rate: base.ScalarOrSchedule,
    *,
    eps: float = 1e-12,
    gamma0: float = 1.0,
    gamma_min: float = 1e-6,
    gamma_max: float = 1e6,
    weight_decay: float = 0.0,
    mask=None,
    strong_wolfe: bool = True,
    wolfe_c1: float = 1e-4,
    wolfe_c2: float = 0.9,
):
    """Spectral-Scaled Broyden (scalar H0) optimizer as an Optax transform.
    def _line_search_strong_wolfe(value_fn, params, f0, g0, p, *, c1=1e-4, c2=0.9, alpha0=1.0, max_iters=20):
        def tree_add(x, a, p):
            return jax.tree_util.tree_map(lambda xi, pi: xi + a * pi, x, p)

        def tree_dot(a, b):
            leaves = jax.tree_util.tree_map(lambda x, y: jnp.vdot(x, y), a, b)
            s = jax.tree_util.tree_reduce(lambda acc, v: acc + v, leaves, 0.0)
            return jnp.real(s) if jnp.iscomplexobj(s) else s

        g0p = tree_dot(g0, p)
        if g0p >= 0:
            p = jax.tree_util.tree_map(lambda x: -x, g0)
            g0p = tree_dot(g0, p)

        alpha = alpha0
        for _ in range(max_iters):
            x_new = tree_add(params, alpha, p)
            f_new = value_fn(x_new)
            if f_new > f0 + c1 * alpha * g0p:
                alpha *= 0.5
                continue
            g_new = jax.grad(value_fn)(x_new)
            gnp = tree_dot(g_new, p)
            if jnp.abs(gnp) <= c2 * jnp.abs(g0p):
                return alpha, p
            if gnp < 0:
                alpha *= 1.1
            else:
                alpha *= 0.5
        return alpha, p


        - Produces updates = -H g with H = gamma I; gamma is BB/Broyden spectral estimate.
        - Chain with learning rate and weight decay.
    """
    lr0 = learning_rate if isinstance(learning_rate, (float, int)) else 1.0
    core = _ssbroyden_core(
        eps=eps,
        gamma0=gamma0,
        gamma_min=gamma_min,
        gamma_max=gamma_max,
        init_lr=float(lr0),
        wolfe=strong_wolfe,
        c1=wolfe_c1,
        c2=wolfe_c2,
    )
    if weight_decay == 0.0:
        return core

    decay = transform.add_decayed_weights(weight_decay, mask)

    def init_fn(params: base.Params) -> tuple[SSBroydenState, base.OptState]:
        return core.init(params), decay.init(params)

    def update_fn(
        updates: base.Updates,
        state: base.OptState,
        params: base.Params | None = None,
        **extra_args: Any,
    ) -> tuple[base.Updates, base.OptState]:
        grads = updates
        core_state, decay_state = _split_decay_state(state)
        updates, core_state = core.update(grads, core_state, params, **extra_args)
        updates, decay_state = decay.update(updates, decay_state, params)
        return updates, (core_state, decay_state)

    return GradientTransformationExtraArgs(init_fn, update_fn)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import coordax as cx
import equinox as eqx
import jax.numpy as jnp

from phydrax._frozendict import frozendict
from phydrax.constraints import enforce_dirichlet
from phydrax.domain import (
    Boundary,
    FixedStart,
    Interval1d,
    PointsBatch,
    ProductStructure,
    TimeInterval,
)
from phydrax.operators.differential import partial_x
from phydrax.solver import (
    EnforcedConstraintPipelines,
    EnforcedInteriorData,
    SingleFieldEnforcedConstraint,
)


def _paired_batch_xp(domain, xs, ps):
    structure = ProductStructure((("x", "p"),)).canonicalize(domain.labels)
    axis_names = structure.axis_names
    assert axis_names is not None
    axis = axis_names[0]
    points = frozendict(
        {
            "x": cx.Field(
                jnp.asarray(xs, dtype=float).reshape((-1, 1)), dims=(axis, None)
            ),
            "p": cx.Field(
                jnp.asarray(ps, dtype=float).reshape((-1, 1)), dims=(axis, None)
            ),
        }
    )
    return PointsBatch(points=points, structure=structure)


def _paired_batch_xpt(domain, xs, ps, ts):
    structure = ProductStructure((("x", "p", "t"),)).canonicalize(domain.labels)
    axis_names = structure.axis_names
    assert axis_names is not None
    axis = axis_names[0]
    points = frozendict(
        {
            "x": cx.Field(
                jnp.asarray(xs, dtype=float).reshape((-1, 1)), dims=(axis, None)
            ),
            "p": cx.Field(
                jnp.asarray(ps, dtype=float).reshape((-1, 1)), dims=(axis, None)
            ),
            "t": cx.Field(jnp.asarray(ts, dtype=float).reshape((-1,)), dims=(axis,)),
        }
    )
    return PointsBatch(points=points, structure=structure)


def test_xp_steady_state_explicit_anchors():
    geom = Interval1d(0.0, 1.0)
    momentum = Interval1d(-1.0, 1.0).relabel("p")
    domain = geom @ momentum

    @domain.Function("x", "p")
    def u(x, p):
        return x[0] + 2.0 * p[0]

    left = domain.component({"x": Boundary()}, where={"x": lambda p: p[0] < 0.5})
    right = domain.component({"x": Boundary()}, where={"x": lambda p: p[0] > 0.5})

    constraints = [
        SingleFieldEnforcedConstraint(
            "u",
            left,
            lambda f: enforce_dirichlet(f, left, var="x", target=1.0),
        ),
        SingleFieldEnforcedConstraint(
            "u",
            right,
            lambda f: enforce_dirichlet(f, right, var="x", target=2.0),
        ),
    ]

    anchors = {
        "x": jnp.array([[0.25], [0.75]], dtype=float),
        "p": jnp.array([[-0.5], [0.5]], dtype=float),
    }
    values = jnp.array([3.0, 4.0], dtype=float)
    interior = EnforcedInteriorData("u", points=anchors, values=values)

    pipelines = EnforcedConstraintPipelines.build(
        functions={"u": u},
        constraints=constraints,
        interior_data=[interior],
        num_reference=128,
    )
    u_enforced = pipelines.apply({"u": u})["u"]
    eval_jit = eqx.filter_jit(lambda b: u_enforced(b).data)

    batch = _paired_batch_xp(
        domain,
        xs=jnp.array([[0.0], [1.0]], dtype=float),
        ps=jnp.array([[0.0], [0.0]], dtype=float),
    )
    out = eval_jit(batch).reshape((-1,))
    assert jnp.allclose(out, jnp.array([1.0, 2.0]), atol=5e-2)

    batch = _paired_batch_xp(domain, xs=anchors["x"], ps=anchors["p"])
    out = eval_jit(batch).reshape((-1,))
    assert jnp.allclose(out, values, atol=1e-3)


def test_xp_coord_separable_partials():
    geom = Interval1d(0.0, 1.0)
    momentum = Interval1d(-1.0, 1.0).relabel("p")
    domain = geom @ momentum

    @domain.Function("x", "p")
    def u(x, p):
        return x[0] ** 2 + 3.0 * p[0]

    component = domain.component()
    sep = component.sample_coord_separable({"x": 6, "p": 5}, num_points=())

    du_dx = partial_x(u, var="x")
    du_dp = partial_x(u, var="p")
    eval_jit = eqx.filter_jit(lambda f, b: f(b).data)

    dx = eval_jit(du_dx, sep)
    dp = eval_jit(du_dp, sep)

    x_axis = sep.points["x"][0].data
    p_axis = sep.points["p"][0].data
    xx, pp = jnp.meshgrid(x_axis, p_axis, indexing="ij")
    assert jnp.allclose(dx, 2.0 * xx, atol=1e-5)
    assert jnp.allclose(dp, 3.0 * jnp.ones_like(pp), atol=1e-5)


def test_xpt_transient_explicit_anchors():
    geom = Interval1d(0.0, 1.0)
    momentum = Interval1d(-1.0, 1.0).relabel("p")
    time = TimeInterval(0.0, 1.0)
    domain = geom @ momentum @ time

    @domain.Function("x", "p", "t")
    def u(x, p, t):
        return x[0] + p[0] + t

    left = domain.component({"x": Boundary()}, where={"x": lambda p: p[0] < 0.5})
    right = domain.component({"x": Boundary()}, where={"x": lambda p: p[0] > 0.5})
    initial = domain.component({"t": FixedStart()})

    constraints = [
        SingleFieldEnforcedConstraint(
            "u",
            left,
            lambda f: enforce_dirichlet(f, left, var="x", target=1.0),
        ),
        SingleFieldEnforcedConstraint(
            "u",
            right,
            lambda f: enforce_dirichlet(f, right, var="x", target=2.0),
        ),
        SingleFieldEnforcedConstraint(
            "u",
            initial,
            lambda f: enforce_dirichlet(f, initial, var="t", target=3.0),
            time_derivative_order=0,
            initial_target=3.0,
        ),
        SingleFieldEnforcedConstraint(
            "u",
            initial,
            lambda f: enforce_dirichlet(f, initial, var="t", target=3.0),
            time_derivative_order=1,
            initial_target=1.0,
        ),
        SingleFieldEnforcedConstraint(
            "u",
            initial,
            lambda f: enforce_dirichlet(f, initial, var="t", target=3.0),
            time_derivative_order=2,
            initial_target=0.0,
        ),
    ]

    anchors = {
        "x": jnp.array([[0.25], [0.75]], dtype=float),
        "p": jnp.array([[-0.5], [0.5]], dtype=float),
        "t": jnp.array([0.4, 0.6], dtype=float),
    }
    values = jnp.array([4.0, 5.0], dtype=float)
    interior = EnforcedInteriorData("u", points=anchors, values=values)

    pipelines = EnforcedConstraintPipelines.build(
        functions={"u": u},
        constraints=constraints,
        interior_data=[interior],
        num_reference=128,
    )
    u_enforced = pipelines.apply({"u": u})["u"]
    eval_jit = eqx.filter_jit(lambda b: u_enforced(b).data)

    batch = _paired_batch_xpt(
        domain,
        xs=jnp.array([[0.0], [1.0]], dtype=float),
        ps=jnp.array([[0.0], [0.0]], dtype=float),
        ts=jnp.array([0.5, 0.5], dtype=float),
    )
    out = eval_jit(batch).reshape((-1,))
    assert jnp.allclose(out, jnp.array([1.0, 2.0]), atol=5e-2)

    batch = _paired_batch_xpt(
        domain,
        xs=jnp.array([[0.5]], dtype=float),
        ps=jnp.array([[0.0]], dtype=float),
        ts=jnp.array([0.0], dtype=float),
    )
    out = eval_jit(batch).reshape((-1,))
    assert jnp.allclose(out, 3.0, atol=2e-2)

    batch = _paired_batch_xpt(domain, xs=anchors["x"], ps=anchors["p"], ts=anchors["t"])
    out = eval_jit(batch).reshape((-1,))
    assert jnp.allclose(out, values, atol=1e-3)


def test_xpt_coord_separable_partials():
    geom = Interval1d(0.0, 1.0)
    momentum = Interval1d(-1.0, 1.0).relabel("p")
    time = TimeInterval(0.0, 1.0)
    domain = geom @ momentum @ time

    @domain.Function("x", "p", "t")
    def u(x, p, t):
        return x[0] ** 2 + 3.0 * p[0] + t

    component = domain.component()
    structure = ProductStructure((("t",),))
    sep = component.sample_coord_separable(
        {"x": 5, "p": 4},
        num_points=6,
        dense_structure=structure,
    )

    du_dx = partial_x(u, var="x")
    du_dp = partial_x(u, var="p")
    eval_jit = eqx.filter_jit(lambda f, b: f(b).data)

    dx = eval_jit(du_dx, sep)
    dp = eval_jit(du_dp, sep)

    t_axis = sep.points["t"].data
    x_axis = sep.points["x"][0].data
    p_axis = sep.points["p"][0].data
    tt, xx, pp = jnp.meshgrid(t_axis, x_axis, p_axis, indexing="ij")
    assert jnp.allclose(dx, 2.0 * xx, atol=1e-5)
    assert jnp.allclose(dp, 3.0 * jnp.ones_like(pp), atol=1e-5)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import coordax as cx
import jax.numpy as jnp

from phydrax._frozendict import frozendict
from phydrax.domain import Interval1d, PointsBatch, ProductStructure, TimeInterval
from phydrax.solver import EnforcedConstraintPipelines, EnforcedInteriorData


def _paired_batch(domain, xs, ts):
    structure = ProductStructure((("x", "t"),)).canonicalize(domain.labels)
    axis_names = structure.axis_names
    assert axis_names is not None
    axis = axis_names[0]
    points = frozendict(
        {
            "x": cx.Field(
                jnp.asarray(xs, dtype=float).reshape((-1, 1)), dims=(axis, None)
            ),
            "t": cx.Field(jnp.asarray(ts, dtype=float).reshape((-1,)), dims=(axis,)),
        }
    )
    return PointsBatch(points=points, structure=structure)


def test_unified_interior_data_tracks_and_scattered():
    geom = Interval1d(0.0, 1.0)
    time = TimeInterval(0.0, 1.0)
    domain = geom @ time

    @domain.Function("x", "t")
    def u(x, t):
        return x[0] * 0.0 + t * 0.0

    scattered = EnforcedInteriorData(
        "u",
        points={
            "x": jnp.array([[0.2], [0.8]], dtype=float),
            "t": jnp.array([0.3, 0.7], dtype=float),
        },
        values=jnp.array([1.0, 2.0], dtype=float),
        use_envelope=True,
        envelope_scale=0.5,
    )

    tracks = EnforcedInteriorData(
        "u",
        sensors=jnp.array([[0.5]], dtype=float),
        times=jnp.array([0.2, 0.6], dtype=float),
        sensor_values=jnp.array([[3.0, 4.0]], dtype=float),
    )

    pipelines = EnforcedConstraintPipelines.build(
        functions={"u": u},
        interior_data=[scattered, tracks],
    )
    u_enforced = pipelines.apply({"u": u})["u"]

    xs = jnp.array([0.2, 0.8, 0.5, 0.5], dtype=float)
    ts = jnp.array([0.3, 0.7, 0.2, 0.6], dtype=float)
    expected = jnp.array([1.0, 2.0, 3.0, 4.0], dtype=float)

    batch = _paired_batch(domain, xs=xs, ts=ts)
    out = jnp.asarray(u_enforced(batch).data).reshape((-1,))
    assert jnp.allclose(out, expected, atol=1e-3)


def test_enforced_interior_data_hermite_track_matches_curve():
    geom = Interval1d(0.0, 1.0)
    time = TimeInterval(0.0, 1.0)
    domain = geom @ time

    @domain.Function("x", "t")
    def u(x, t):
        return x[0] * 0.0 + t * 0.0

    tracks = EnforcedInteriorData(
        "u",
        sensors=jnp.array([[0.5]], dtype=float),
        times=jnp.array([0.0, 1.0], dtype=float),
        sensor_values=jnp.array([[0.0, 1.0]], dtype=float),
        time_interp="hermite",
    )

    pipelines = EnforcedConstraintPipelines.build(
        functions={"u": u},
        interior_data=[tracks],
    )
    u_enforced = pipelines.apply({"u": u})["u"]

    xs = jnp.array([0.5, 0.5], dtype=float)
    ts = jnp.array([0.0, 0.5], dtype=float)
    expected = jnp.array([0.0, 0.5], dtype=float)

    batch = _paired_batch(domain, xs=xs, ts=ts)
    out = jnp.asarray(u_enforced(batch).data).reshape((-1,))
    assert jnp.allclose(out, expected, atol=1e-3)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

"""Optimization algorithms and Optax gradient transformations."""

from ._bfgs_sw import bfgs_sw
from ._lbfgs_sw import lbfgs_sw
from ._ssbroyden import ssbroyden


__all__ = [
    "bfgs_sw",
    "lbfgs_sw",
    "ssbroyden",
]



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

from __future__ import annotations

from collections.abc import Callable
from typing import Any

import jax.numpy as jnp

from .._callable import _ensure_special_kwonly_args
from .._strict import StrictModule
from ..nn.models.core._base import _AbstractBaseModel


class StructuredCallable(StrictModule):
    """Wrapper marking a callable as accepting structured (tuple) inputs."""

    func: Callable

    def __init__(self, func: Callable, /):
        self.func = _ensure_special_kwonly_args(func)

    def __call__(self, x: Any, /, *, key=None, iter_=None, **kwargs: Any):
        return self.func(x, key=key, iter_=iter_, **kwargs)


def structured(func: Callable, /) -> StructuredCallable:
    """Mark a callable as supporting structured (tuple) inputs."""
    return StructuredCallable(func)


class _ConcatenatedModelCallable(StrictModule):
    model: Callable
    supports_structured_input: bool

    def __init__(self, model: Callable, /):
        supports_structured_input = isinstance(model, StructuredCallable)
        if isinstance(model, _AbstractBaseModel) and model.supports_structured_input():
            supports_structured_input = True
        self.supports_structured_input = bool(supports_structured_input)
        self.model = _ensure_special_kwonly_args(model)

    def __call__(self, *args: Any, key=None, iter_=None, **kwargs: Any):
        if not args:
            raise ValueError("Model callable requires at least one positional input.")

        if self.supports_structured_input:
            if len(args) == 1:
                x_in = args[0]
            else:
                flat: list[Any] = []
                for a in args:
                    if isinstance(a, tuple):
                        flat.extend(a)
                    else:
                        flat.append(a)
                x_in = tuple(jnp.asarray(x) for x in flat)
            return self.model(x_in, key=key, iter_=iter_, **kwargs)

        for a in args:
            if isinstance(a, tuple):
                raise ValueError(
                    "Model callable does not support structured inputs; got a tuple argument. "
                    "Use a model that supports_structured_input() or explicitly materialize the grid."
                )
            arr = jnp.asarray(a)
            if arr.ndim > 1:
                raise ValueError(
                    "Model callable does not support batched/structured inputs; got input with shape "
                    f"{arr.shape}. Use a model that supports_structured_input() or explicitly materialize the grid."
                )

        parts = [jnp.asarray(x).reshape((-1,)) for x in args]
        x_in = parts[0] if len(parts) == 1 else jnp.concatenate(parts, axis=0)
        return self.model(x_in, key=key, iter_=iter_, **kwargs)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import coordax as cx
import jax.numpy as jnp

from phydrax._frozendict import frozendict
from phydrax.domain import Cube, DomainFunction, Square, TimeInterval
from phydrax.operators.linalg import det


def test_det_simple_matrix_function():
    geom = Square(center=(0.0, 0.0), side=2.0)

    @geom.Function("x")
    def u(x):
        return jnp.array([[x[0], 0], [0, x[1]]])

    det_u = det(u)
    pts = frozendict({"x": cx.Field(jnp.array([2.0, 3.0]), dims=(None,))})
    result = jnp.asarray(det_u(pts).data)

    expected = 6.0  # 2*3
    assert jnp.allclose(result, expected)


def test_det_time_dependent_matrix_function():
    dom = Square(center=(0.0, 0.0), side=2.0) @ TimeInterval(0.0, 1.0)

    @dom.Function("x", "t")
    def u(x, t):
        return jnp.array([[x[0] * t, 0.0], [0.0, x[1] * t]])

    det_u = det(u)
    pts = frozendict(
        {
            "x": cx.Field(jnp.array([2.0, 3.0]), dims=(None,)),
            "t": cx.Field(jnp.array(0.5), dims=()),
        }
    )
    result = jnp.asarray(det_u(pts).data)

    expected = 1.5  # 2*3*0.5^2
    assert jnp.allclose(result, expected)


def test_det_complex_function():
    geom = Square(center=(0.0, 0.0), side=2.0)

    @geom.Function("x")
    def u(x):
        return jnp.array([[x[0], 0], [0, 1j * x[1]]])

    det_u = det(u)
    pts = frozendict({"x": cx.Field(jnp.array([2.0, 3.0]), dims=(None,))})
    result = jnp.asarray(det_u(pts).data)

    expected = 6.0j  # i*2*3
    assert jnp.allclose(result, expected)


def test_det_non_diagonal_matrix():
    geom = Square(center=(0.0, 0.0), side=2.0)

    @geom.Function("x")
    def u(x):
        return jnp.array([[x[0], x[1]], [x[1], x[0]]])

    det_u = det(u)
    pts = frozendict({"x": cx.Field(jnp.array([3.0, 2.0]), dims=(None,))})
    result = jnp.asarray(det_u(pts).data)

    expected = 5.0  # 3*3 - 2*2 = 9 - 4 = 5
    assert jnp.allclose(result, expected)


def test_det_3x3_matrix():
    geom = Cube(center=(0.0, 0.0, 0.0), side=2.0)

    @geom.Function("x")
    def u(x):
        return jnp.array([[x[0], x[1], x[2]], [x[2], x[0], x[1]], [x[1], x[2], x[0]]])

    det_u = det(u)
    pts = frozendict({"x": cx.Field(jnp.array([1.0, 2.0, 3.0]), dims=(None,))})
    result = jnp.asarray(det_u(pts).data)

    # For the matrix [[1, 2, 3], [3, 1, 2], [2, 3, 1]], the determinant is 18.
    expected = 18.0
    assert jnp.allclose(result, expected)


def test_det_preserves_metadata():
    geom = Square(center=(0.0, 0.0), side=2.0)
    u = DomainFunction(
        domain=geom, deps=("x",), func=lambda x: jnp.eye(2), metadata={"tag": "keep"}
    )
    out = det(u)
    assert out.metadata == u.metadata



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import jax.numpy as jnp
import jax.random as jr

from phydrax.domain import DatasetDomain, FourierAxisSpec, Interval1d, ProductStructure
from phydrax.operators.integral import integral


def test_dataset_domain_samples_points_batch():
    data = jnp.arange(10.0, dtype=float).reshape((10, 1))
    dom = DatasetDomain(data)
    component = dom.component()
    structure = ProductStructure((("data",),))

    batch = component.sample(4, structure=structure, key=jr.key(0))
    axis = batch.structure.axis_for("data")
    assert axis is not None

    field = batch["data"]
    assert field.dims == (axis, None)
    assert field.data.shape == (4, 1)


def test_dataset_domain_integral_probability_measure_is_average():
    data = jnp.zeros((5, 2), dtype=float)
    dom = DatasetDomain(data, measure="probability")
    component = dom.component()
    structure = ProductStructure((("data",),))

    batch = component.sample(3, structure=structure, key=jr.key(0))
    u = dom.Function()(1.0)
    out = integral(u, batch, component=component)
    assert jnp.allclose(jnp.asarray(out.data), 1.0)


def test_dataset_domain_integral_count_measure_is_sum():
    data = jnp.zeros((5, 2), dtype=float)
    dom = DatasetDomain(data, measure="count")
    component = dom.component()
    structure = ProductStructure((("data",),))

    batch = component.sample(3, structure=structure, key=jr.key(0))
    u = dom.Function()(1.0)
    out = integral(u, batch, component=component)
    assert jnp.allclose(jnp.asarray(out.data), 5.0)


def test_dataset_domain_with_coord_separable_geometry_sampling():
    data = jnp.arange(6.0, dtype=float)
    data_dom = DatasetDomain(data)
    geom = Interval1d(0.0, 1.0)
    domain = data_dom @ geom

    component = domain.component()
    dense_structure = ProductStructure((("data",),))
    batch = component.sample_coord_separable(
        {"x": FourierAxisSpec(8)},
        num_points=3,
        dense_structure=dense_structure,
        key=jr.key(0),
    )

    axis = batch.dense_structure.axis_for("data")
    assert axis is not None
    assert batch["data"].dims == (axis,)
    assert batch["data"].data.shape == (3,)
    assert isinstance(batch["x"], tuple)
    assert batch["x"][0].data.shape == (8,)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

from __future__ import annotations

import dataclasses

import jax
import jax.numpy as jnp
import numpy as np
from jaxtyping import Array, ArrayLike


@dataclasses.dataclass(frozen=True)
class PackedBVH:
    """A packed binary AABB BVH with fixed-size leaf payloads."""

    bbox_min: Array  # (nNodes, dim)
    bbox_max: Array  # (nNodes, dim)
    left: Array  # (nNodes,) int32, -1 for leaf
    right: Array  # (nNodes,) int32, -1 for leaf
    leaf_id: Array  # (nNodes,) int32, >=0 for leaf else -1

    leaf_items: Array  # (nLeaves, leaf_size) int32, padded with -1
    leaf_bbox_min: Array  # (nLeaves, dim)
    leaf_bbox_max: Array  # (nLeaves, dim)

    leaf_size: int
    max_depth: int


def build_packed_bvh(
    item_bbox_min: np.ndarray,
    item_bbox_max: np.ndarray,
    centers: np.ndarray | None = None,
    /,
    *,
    leaf_size: int = 16,
    dtype: jnp.dtype = jnp.float32,
) -> PackedBVH:
    """Build a packed BVH over axis-aligned item bounding boxes (NumPy build)."""
    bmin = np.asarray(item_bbox_min, dtype=np.float64)
    bmax = np.asarray(item_bbox_max, dtype=np.float64)
    if bmin.ndim != 2 or bmax.ndim != 2:
        raise ValueError("item_bbox_min/max must be rank-2 arrays.")
    if bmin.shape != bmax.shape:
        raise ValueError("item_bbox_min/max must have matching shapes.")
    if bmin.shape[0] == 0:
        raise ValueError("cannot build BVH over empty item set.")
    if leaf_size <= 0:
        raise ValueError(f"leaf_size must be positive, got {leaf_size}.")

    if centers is None:
        ctr = 0.5 * (bmin + bmax)
    else:
        ctr = np.asarray(centers, dtype=np.float64)
        if ctr.shape != bmin.shape:
            raise ValueError("centers must have shape (nItems, dim).")

    bbox_min_nodes: list[np.ndarray] = []
    bbox_max_nodes: list[np.ndarray] = []
    left_nodes: list[int] = []
    right_nodes: list[int] = []
    leaf_id_nodes: list[int] = []
    leaf_items_list: list[np.ndarray] = []
    leaf_node_for_id: list[int] = []

    def _build(indices: np.ndarray, depth: int) -> tuple[int, int]:
        node = len(left_nodes)
        left_nodes.append(-1)
        right_nodes.append(-1)
        leaf_id_nodes.append(-1)

        bmin_n = bmin[indices].min(axis=0)
        bmax_n = bmax[indices].max(axis=0)
        bbox_min_nodes.append(bmin_n)
        bbox_max_nodes.append(bmax_n)

        max_depth = depth
        if indices.size <= leaf_size:
            lid = len(leaf_items_list)
            leaf = np.full((leaf_size,), -1, dtype=np.int32)
            leaf[: indices.size] = indices.astype(np.int32, copy=False)
            leaf_items_list.append(leaf)
            leaf_id_nodes[node] = lid
            leaf_node_for_id.append(node)
            return node, max_depth

        extent = bmax_n - bmin_n
        axis = int(np.argmax(extent))
        vals = ctr[indices, axis]
        mid = int(indices.size // 2)
        part = np.argpartition(vals, mid)
        left_idx = indices[part[:mid]]
        right_idx = indices[part[mid:]]
        if left_idx.size == 0 or right_idx.size == 0:
            left_idx = indices[:mid]
            right_idx = indices[mid:]

        lnode, ldepth = _build(left_idx, depth + 1)
        rnode, rdepth = _build(right_idx, depth + 1)
        left_nodes[node] = int(lnode)
        right_nodes[node] = int(rnode)
        max_depth = max(max_depth, ldepth, rdepth)
        return node, max_depth

    n_items = int(bmin.shape[0])
    root, max_depth = _build(np.arange(n_items, dtype=np.int32), 0)
    if root != 0:
        raise RuntimeError("BVH build invariant violated: root must be node 0.")

    bbox_min_np = np.stack(bbox_min_nodes, axis=0)
    bbox_max_np = np.stack(bbox_max_nodes, axis=0)
    left_np = np.asarray(left_nodes, dtype=np.int32)
    right_np = np.asarray(right_nodes, dtype=np.int32)
    leaf_id_np = np.asarray(leaf_id_nodes, dtype=np.int32)
    leaf_items_np = np.stack(leaf_items_list, axis=0).astype(np.int32, copy=False)

    leaf_node_for_id_np = np.asarray(leaf_node_for_id, dtype=np.int32)
    leaf_bbox_min_np = bbox_min_np[leaf_node_for_id_np]
    leaf_bbox_max_np = bbox_max_np[leaf_node_for_id_np]

    return PackedBVH(
        bbox_min=jnp.asarray(bbox_min_np, dtype=dtype),
        bbox_max=jnp.asarray(bbox_max_np, dtype=dtype),
        left=jnp.asarray(left_np, dtype=jnp.int32),
        right=jnp.asarray(right_np, dtype=jnp.int32),
        leaf_id=jnp.asarray(leaf_id_np, dtype=jnp.int32),
        leaf_items=jnp.asarray(leaf_items_np, dtype=jnp.int32),
        leaf_bbox_min=jnp.asarray(leaf_bbox_min_np, dtype=dtype),
        leaf_bbox_max=jnp.asarray(leaf_bbox_max_np, dtype=dtype),
        leaf_size=int(leaf_size),
        max_depth=int(max_depth),
    )


def build_point_bvh(
    points: ArrayLike,
    /,
    *,
    leaf_size: int = 32,
    dtype: jnp.dtype = jnp.float32,
) -> PackedBVH:
    pts = np.asarray(points, dtype=np.float64)
    if pts.ndim != 2:
        raise ValueError("points must have shape (nPoints, dim).")
    return build_packed_bvh(
        pts,
        pts,
        pts,
        leaf_size=leaf_size,
        dtype=dtype,
    )


def aabb_dist2(p: Array, bmin: Array, bmax: Array, /) -> Array:
    z = jnp.asarray(0.0, dtype=p.dtype)
    d = jnp.maximum(z, jnp.maximum(bmin - p, p - bmax))
    return jnp.sum(d * d, axis=-1)


def beam_select_nodes(
    points: Array,
    /,
    *,
    bvh: PackedBVH,
    beam_width: int,
    steps: int,
) -> Array:
    """Beam traverse BVH by AABB lower bounds. Returns node ids with shape (N, B)."""
    if beam_width <= 0:
        raise ValueError(f"beam_width must be positive, got {beam_width}.")
    if steps <= 0:
        raise ValueError(f"steps must be positive, got {steps}.")

    pts = jnp.asarray(points, dtype=bvh.bbox_min.dtype)
    if pts.ndim == 1:
        pts = pts.reshape((1, -1))
    if pts.ndim != 2:
        raise ValueError("points must have shape (N, dim) or (dim,).")

    B = int(beam_width)
    inf = jnp.asarray(jnp.inf, dtype=pts.dtype)

    bbox_min = bvh.bbox_min
    bbox_max = bvh.bbox_max
    left = bvh.left
    right = bvh.right
    leaf_id = bvh.leaf_id

    nodes = jnp.full((pts.shape[0], B), jnp.int32(-1))
    nodes = nodes.at[:, 0].set(jnp.int32(0))

    def _step(_, nodes):
        valid = nodes >= 0
        safe = jnp.where(valid, nodes, jnp.int32(0))
        is_leaf = valid & (leaf_id[safe] >= 0)

        lch = jnp.where(valid, left[safe], jnp.int32(-1))
        rch = jnp.where(valid, right[safe], jnp.int32(-1))

        cand0 = jnp.where(is_leaf, safe, lch)
        cand1 = jnp.where(is_leaf, jnp.int32(-1), rch)
        cand_nodes = jnp.concatenate([cand0, cand1], axis=1)  # (N, 2B)

        cand_safe = jnp.where(cand_nodes >= 0, cand_nodes, jnp.int32(0))
        bmin = bbox_min[cand_safe]
        bmax = bbox_max[cand_safe]
        d2 = aabb_dist2(pts[:, None, :], bmin, bmax)
        d2 = jnp.where(cand_nodes >= 0, d2, inf)

        _, idx = jax.lax.top_k(-d2, B)
        return jnp.take_along_axis(cand_nodes, idx, axis=1)

    nodes = jax.lax.fori_loop(0, int(steps), _step, nodes)
    return nodes


def beam_select_leaf_items(
    points: Array,
    /,
    *,
    bvh: PackedBVH,
    beam_width: int,
    steps: int,
) -> tuple[Array, Array]:
    """Return candidate leaf items (and validity mask) using beam BVH traversal."""
    pts = jnp.asarray(points, dtype=bvh.bbox_min.dtype)
    is_single = pts.ndim == 1
    if is_single:
        pts = pts.reshape((1, -1))
    if pts.ndim != 2:
        raise ValueError("points must have shape (N, dim) or (dim,).")

    nodes = beam_select_nodes(pts, bvh=bvh, beam_width=beam_width, steps=steps)

    safe_nodes = jnp.where(nodes >= 0, nodes, jnp.int32(0))
    lids = bvh.leaf_id[safe_nodes]
    valid_leaf = (nodes >= 0) & (lids >= 0)
    safe_lids = jnp.where(valid_leaf, lids, jnp.int32(0))

    items = bvh.leaf_items[safe_lids]  # (N, B, leaf_size)
    items = jnp.where(valid_leaf[..., None], items, jnp.int32(-1))
    items = items.reshape((pts.shape[0], int(beam_width * bvh.leaf_size)))

    valid = items >= 0
    safe_items = jnp.where(valid, items, jnp.int32(0))
    if is_single:
        return safe_items.reshape((-1,)), valid.reshape((-1,))
    return safe_items, valid


__all__ = [
    "PackedBVH",
    "aabb_dist2",
    "beam_select_leaf_items",
    "beam_select_nodes",
    "build_packed_bvh",
    "build_point_bvh",
]



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import inspect
from collections.abc import Callable
from typing import Any


def _compile_ctx_integrand(integrand: Callable, /) -> Callable[[dict[str, Any]], Any]:
    """Compile a ctx-dict integrand dispatcher.

    The returned callable takes a context dict and calls `integrand` with only the
    parameters it declares, in signature order.
    """
    arg_names = tuple(inspect.signature(integrand).parameters.keys())

    def call(ctx: dict[str, Any]):
        return integrand(*tuple(ctx[name] for name in arg_names))

    return call



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

from __future__ import annotations

from collections.abc import Callable, Mapping, Sequence
from typing import Any, Literal

import coordax as cx
import jax.numpy as jnp
from jaxtyping import Array, ArrayLike, Key

from .._doc import DOC_KEY0
from .._strict import StrictModule
from ..domain._components import (
    DomainComponent,
    DomainComponentUnion,
    Fixed,
    FixedEnd,
    FixedStart,
    Interior,
)
from ..domain._function import DomainFunction
from ..domain._structure import (
    CoordSeparableBatch,
    NumPoints,
    PointsBatch,
    ProductStructure,
)
from ..operators.integral._batch_ops import integral, mean
from ._base import AbstractSamplingConstraint


class _SquaredFrobeniusResidual(StrictModule):
    residual: DomainFunction

    def __init__(self, residual: DomainFunction):
        self.residual = residual

    def __call__(self, *args: Any, key=None, **kwargs: Any):
        y = jnp.asarray(self.residual.func(*args, key=key, **kwargs))
        return jnp.sum(y * y)


class FunctionalConstraint(AbstractSamplingConstraint):
    r"""A sampled objective term defined by a residual `DomainFunction`.

    A `FunctionalConstraint` represents one term in a physics/data objective. It is
    defined by:

    - a `DomainComponent` (or union) describing the integration/sampling region
      $\Omega_{\text{comp}}$ and measure $\mu$;
    - a residual operator producing a `DomainFunction` $r(z)$ from the current set of
      field functions.

    The pointwise squared residual is taken as a Frobenius norm:

    $$
    \rho(z) = \|r(z)\|_F^2 = \sum_{i} r_i(z)^2,
    $$

    and the scalar loss is computed using either reduction mode.

    For `reduction="mean"`:

    $$
    \ell = w\,\frac{1}{\mu(\Omega_{\text{comp}})}\int_{\Omega_{\text{comp}}} \rho(z)\,d\mu(z),
    $$

    For `reduction="integral"`:

    $$
    \ell = w\int_{\Omega_{\text{comp}}} \rho(z)\,d\mu(z),
    $$

    where $w$ is the scalar `weight`.

    Sampling is performed according to `structure` (paired blocks) and optionally
    `coord_separable`.
    """

    constraint_vars: tuple[str, ...]
    component: DomainComponent | DomainComponentUnion
    structure: ProductStructure
    coord_separable: Mapping[str, Any] | None
    dense_structure: ProductStructure | None
    num_points: NumPoints | tuple[Any, ...]
    sampler: str
    weight: Array
    label: str | None
    over: str | tuple[str, ...] | None
    reduction: Literal["mean", "integral"]
    residual: Callable[[Mapping[str, DomainFunction]], DomainFunction]

    def __init__(
        self,
        *,
        component: DomainComponent | DomainComponentUnion,
        residual: Callable[[Mapping[str, DomainFunction]], DomainFunction],
        num_points: NumPoints | tuple[Any, ...],
        structure: ProductStructure,
        coord_separable: Mapping[str, Any] | None = None,
        dense_structure: ProductStructure | None = None,
        constraint_vars: Sequence[str] | None = None,
        sampler: str = "latin_hypercube",
        weight: ArrayLike = 1.0,
        label: str | None = None,
        over: str | tuple[str, ...] | None = None,
        reduction: Literal["mean", "integral"] = "mean",
    ):
        self.constraint_vars = () if constraint_vars is None else tuple(constraint_vars)
        self.component = component
        self.residual = residual
        self.num_points = num_points
        self.structure = structure
        self.coord_separable = coord_separable
        self.dense_structure = dense_structure
        self.sampler = str(sampler)
        self.weight = jnp.asarray(weight, dtype=float)
        self.label = None if label is None else str(label)
        self.over = over
        self.reduction = reduction
        self._ensure_coord_separable_interior()

    def _ensure_coord_separable_interior(self) -> None:
        if self.coord_separable is None:
            return
        if isinstance(self.component, DomainComponentUnion):
            raise ValueError(
                "coord_separable sampling is not supported for DomainComponentUnion."
            )
        coord_labels = tuple(self.coord_separable)
        bad = tuple(
            lbl
            for lbl in coord_labels
            if not isinstance(self.component.spec.component_for(lbl), Interior)
        )
        if bad:
            raise ValueError(
                f"coord_separable labels must use Interior() components; got {bad!r}."
            )

    @classmethod
    def from_operator(
        cls,
        *,
        component: DomainComponent | DomainComponentUnion,
        operator: Callable[..., DomainFunction],
        constraint_vars: str | Sequence[str],
        num_points: NumPoints | tuple[Any, ...],
        structure: ProductStructure,
        coord_separable: Mapping[str, int | Sequence[int]] | None = None,
        dense_structure: ProductStructure | None = None,
        sampler: str = "latin_hypercube",
        weight: ArrayLike = 1.0,
        label: str | None = None,
        over: str | tuple[str, ...] | None = None,
        reduction: Literal["mean", "integral"] = "mean",
    ) -> "FunctionalConstraint":
        r"""Create a `FunctionalConstraint` from an operator mapping `DomainFunction`s to a residual.

        This wraps an `operator(u1, u2, ...) -> r` into a residual callable
        `residual(functions) -> r` using the provided `constraint_vars`.
        """
        vars_tuple = (
            (constraint_vars,)
            if isinstance(constraint_vars, str)
            else tuple(constraint_vars)
        )

        def residual(functions: Mapping[str, DomainFunction], /) -> DomainFunction:
            return operator(*(functions[name] for name in vars_tuple))

        return cls(
            component=component,
            residual=residual,
            num_points=num_points,
            structure=structure,
            coord_separable=coord_separable,
            dense_structure=dense_structure,
            constraint_vars=vars_tuple,
            sampler=sampler,
            weight=weight,
            label=label,
            over=over,
            reduction=reduction,
        )

    def _dense_structure_for_coord_separable(self) -> ProductStructure | None:
        if self.dense_structure is not None:
            return self.dense_structure
        if self.coord_separable is None:
            return None
        if isinstance(self.component, DomainComponentUnion):
            raise ValueError(
                "coord_separable sampling is not supported for DomainComponentUnion."
            )
        coord_labels = set(self.coord_separable)
        fixed_labels = frozenset(
            lbl
            for lbl in self.component.domain.labels
            if isinstance(
                self.component.spec.component_for(lbl), (FixedStart, FixedEnd, Fixed)
            )
        )
        dense_labels = [
            lbl
            for lbl in self.component.domain.labels
            if lbl not in coord_labels and lbl not in fixed_labels
        ]
        if not dense_labels:
            return ProductStructure(blocks=())
        blocks: list[tuple[str, ...]] = []
        for block in self.structure.blocks:
            filtered = tuple(
                lbl
                for lbl in block
                if lbl not in coord_labels and lbl not in fixed_labels
            )
            if filtered:
                blocks.append(filtered)
        covered = set(lbl for block in blocks for lbl in block)
        missing = [lbl for lbl in dense_labels if lbl not in covered]
        if missing:
            raise ValueError(
                "coord_separable requires dense_structure to cover non-separable labels; "
                f"missing {tuple(missing)!r}."
            )
        return ProductStructure(tuple(blocks))

    def sample(
        self,
        *,
        key: Key[Array, ""] = DOC_KEY0,
    ) -> PointsBatch | CoordSeparableBatch | tuple[PointsBatch, ...]:
        r"""Sample points from the configured component.

        - Returns a `PointsBatch` for paired sampling (`coord_separable=None`).
        - Returns a `CoordSeparableBatch` when `coord_separable` is enabled.
        - Returns a tuple of `PointsBatch` when sampling from a `DomainComponentUnion`.
        """
        if self.coord_separable is not None:
            if isinstance(self.component, DomainComponentUnion):
                raise ValueError(
                    "coord_separable sampling is not supported for DomainComponentUnion."
                )
            return self.component.sample_coord_separable(
                self.coord_separable,
                num_points=self.num_points,
                dense_structure=self._dense_structure_for_coord_separable(),
                sampler=self.sampler,
                key=key,
            )
        return self.component.sample(
            self.num_points,
            structure=self.structure,
            sampler=self.sampler,
            key=key,
        )

    def loss(
        self,
        functions: Mapping[str, DomainFunction],
        /,
        *,
        key: Key[Array, ""] = DOC_KEY0,
        batch: PointsBatch | CoordSeparableBatch | tuple[PointsBatch, ...] | None = None,
        **kwargs: Any,
    ) -> Array:
        r"""Evaluate the scalar loss for this constraint.

        This samples the configured component, evaluates the residual, forms a squared
        Frobenius norm, and reduces via `mean(...)` or `integral(...)` depending on
        `reduction` and `over`.

        This:
        1) builds the residual `DomainFunction` $r$ from `functions`,
        2) samples points $z_i$ on `component`,
        3) computes $\rho(z_i)=\|r(z_i)\|_F^2$,
        4) reduces using either a mean or an integral estimator.
        """
        res = self.residual(functions)
        if not isinstance(res, DomainFunction):
            base = None
            if self.constraint_vars:
                base = functions.get(self.constraint_vars[0])
            if base is None:
                for fn in functions.values():
                    if isinstance(fn, DomainFunction):
                        base = fn
                        break
            domain = base.domain if base is not None else self.component.domain
            if callable(res):
                deps = base.deps if base is not None else domain.labels
                res = DomainFunction(domain=domain, deps=deps, func=res, metadata={})
            else:
                res = DomainFunction(domain=domain, deps=(), func=res, metadata={})

        batch_ = self.sample(key=key) if batch is None else batch
        f = DomainFunction(
            domain=res.domain,
            deps=res.deps,
            func=_SquaredFrobeniusResidual(res),
            metadata=res.metadata,
        )
        if self.reduction == "mean":
            out = mean(
                f,
                batch_,
                component=self.component,
                over=self.over,
                key=key,
                **kwargs,
            )
        else:
            out = integral(
                f,
                batch_,
                component=self.component,
                over=self.over,
                key=key,
                **kwargs,
            )
        if not isinstance(out, cx.Field):
            raise TypeError("Expected reduction to return a coordax.Field.")
        if out.dims != ():
            raise ValueError(
                f"Constraint reduction must return a scalar Field, got dims={out.dims}."
            )
        return self.weight * jnp.asarray(out.data, dtype=float).reshape(())



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import jax
import jax.numpy as jnp
import jax.random as jr

from phydrax.nn.models import FeynmaNN, LatentContractionModel, MLP, Separable


def test_pimlp_keep_output_complex_scalar():
    key = jr.key(0)
    m = FeynmaNN(
        in_size=2,
        out_size="scalar",
        width_size=16,
        depth=2,
        num_paths=3,
        keep_output_complex=True,
        key=key,
    )
    x = jr.normal(jr.key(1), (2,))
    y = m(x)
    assert jnp.iscomplexobj(y)
    assert y.shape == ()


def test_pimlp_keep_output_complex_vector():
    m = FeynmaNN(
        in_size=3,
        out_size=4,
        width_size=16,
        depth=2,
        num_paths=3,
        keep_output_complex=True,
        key=jr.key(2),
    )
    x = jr.normal(jr.key(3), (3,))
    y = m(x)
    assert jnp.iscomplexobj(y)
    assert y.shape == (4,)


def test_separable_wrapper_complex_output_toggle():
    # Underlying scalar models for each coordinate produce complex latents
    L = 2
    out = 1
    models = [
        FeynmaNN(
            in_size="scalar",
            out_size=L * out,
            width_size=12,
            depth=2,
            num_paths=2,
            keep_output_complex=True,
            key=jr.key(10 + i),
        )
        for i in range(2)
    ]
    sep_c = Separable(
        in_size=2,
        out_size="scalar",
        latent_size=L,
        models=models,
        keep_outputs_complex=True,
    )
    x = jr.normal(jr.key(12), (7, 2))
    yc = jax.vmap(sep_c)(x)
    assert jnp.iscomplexobj(yc)
    assert yc.shape == (7,)

    sep_r = Separable(
        in_size=2,
        out_size="scalar",
        latent_size=L,
        models=models,
        keep_outputs_complex=False,
    )
    yr = jax.vmap(sep_r)(x)
    assert not jnp.iscomplexobj(yr)
    assert yr.shape == (7,)


def test_latent_contraction_complex_output_toggle():
    # Space model produces complex latents of size L*out; time model is real L
    L = 3
    d = 2
    space_model = FeynmaNN(
        in_size=d,
        out_size=L * 1,
        width_size=16,
        depth=2,
        num_paths=2,
        keep_output_complex=True,
        key=jr.key(20),
    )
    time_model = MLP(in_size="scalar", out_size=L, width_size=8, depth=1, key=jr.key(21))
    ts_c = LatentContractionModel(
        out_size="scalar",
        latent_size=L,
        x=space_model,
        t=time_model,
        keep_outputs_complex=True,
    )
    Ns, Nt = 4, 5
    x_space = jr.normal(jr.key(22), (Ns, d))
    x_time = jr.normal(jr.key(23), (Nt,))
    space_rep = jnp.repeat(x_space, Nt, axis=0)
    time_rep = jnp.tile(x_time, Ns)
    points = jnp.column_stack([space_rep, time_rep])
    yc = jax.vmap(ts_c)(points).reshape(Ns, Nt)
    assert jnp.iscomplexobj(yc)
    assert yc.shape == (Ns, Nt)

    ts_r = LatentContractionModel(
        out_size="scalar",
        latent_size=L,
        x=space_model,
        t=time_model,
        keep_outputs_complex=False,
    )
    yr = jax.vmap(ts_r)(points).reshape(Ns, Nt)
    assert not jnp.iscomplexobj(yr)
    assert yr.shape == (Ns, Nt)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import coordax as cx
import equinox as eqx
import jax.numpy as jnp
import jax.random as jr

from phydrax._frozendict import frozendict
from phydrax.constraints import ContinuousInitialConstraint, enforce_dirichlet
from phydrax.domain import (
    Boundary,
    FixedStart,
    Interval1d,
    PointsBatch,
    ProductStructure,
    TimeInterval,
)
from phydrax.domain._structure import CoordSeparableBatch
from phydrax.solver import (
    EnforcedInteriorData,
    FunctionalSolver,
    SingleFieldEnforcedConstraint,
)


def _paired_batch(domain, xs, ts):
    structure = ProductStructure((("x", "t"),)).canonicalize(domain.labels)
    axis_names = structure.axis_names
    assert axis_names is not None
    axis = axis_names[0]
    points = frozendict(
        {
            "x": cx.Field(
                jnp.asarray(xs, dtype=float).reshape((-1, 1)), dims=(axis, None)
            ),
            "t": cx.Field(jnp.asarray(ts, dtype=float).reshape((-1,)), dims=(axis,)),
        }
    )
    return PointsBatch(points=points, structure=structure)


def test_functional_solver_builds_enforced_pipeline_terms():
    geom = Interval1d(0.0, 1.0)
    time = TimeInterval(0.0, 1.0)
    domain = geom @ time

    @domain.Function("x", "t")
    def u(x, t):
        return x[0] + t

    boundary_component = domain.component({"x": Boundary()})
    initial_component = domain.component({"t": FixedStart()})

    boundary_constraint = SingleFieldEnforcedConstraint(
        "u",
        boundary_component,
        lambda f: enforce_dirichlet(f, boundary_component, var="x", target=5.0),
    )
    initial_constraint = SingleFieldEnforcedConstraint(
        "u",
        initial_component,
        lambda f: enforce_dirichlet(f, initial_component, var="t", target=2.0),
    )

    anchors = {
        "x": jnp.array([[0.25], [0.75]], dtype=float),
        "t": jnp.array([0.6, 0.4], dtype=float),
    }
    anchor_values = jnp.array([3.0, 4.0], dtype=float)
    interior = EnforcedInteriorData(
        "u",
        points=anchors,
        values=anchor_values,
        eps_snap=1e-12,
    )

    solver = FunctionalSolver(
        functions={"u": u},
        constraints=(),
        constraint_terms=[boundary_constraint, initial_constraint],
        interior_data_terms=[interior],
        boundary_weight_num_reference=256,
    )
    u_enforced = solver.ansatz_functions()["u"]
    eval_jit = eqx.filter_jit(lambda b: u_enforced(b).data)

    out = eval_jit(
        _paired_batch(domain, xs=jnp.array([0.0, 1.0]), ts=jnp.array([0.5, 0.5]))
    )
    assert jnp.allclose(out.reshape((-1,)), 5.0, atol=1e-3)

    out = eval_jit(
        _paired_batch(domain, xs=jnp.array([0.5, 0.5]), ts=jnp.array([0.0, 0.0]))
    )
    assert jnp.allclose(out.reshape((-1,)), 2.0, atol=1e-3)

    out = eval_jit(
        _paired_batch(domain, xs=jnp.array([0.25, 0.75]), ts=jnp.array([0.6, 0.4]))
    )
    assert jnp.allclose(out.reshape((-1,)), anchor_values, atol=1e-3)


def test_initial_constraint_coord_separable_spatial():
    geom = Interval1d(0.0, 1.0)
    time = TimeInterval(0.0, 1.0)
    domain = geom @ time

    @domain.Function("x", "t")
    def u(x, t):
        return 0.0

    initial_component = domain.component({"t": FixedStart()})
    structure = ProductStructure((("x",),))

    constraint = ContinuousInitialConstraint(
        "u",
        initial_component,
        func=0.0,
        num_points=(),
        structure=structure,
        coord_separable={"x": 5},
    )

    batch = constraint.sample(key=jr.key(0))
    assert isinstance(batch, CoordSeparableBatch)
    assert isinstance(batch.points["x"], tuple)
    assert len(batch.points["x"]) == 1

    loss = constraint.loss({"u": u}, key=jr.key(0))
    assert jnp.allclose(loss, 0.0, atol=1e-6)



#
#  Copyright © 2026 PHYDRA, Inc. All rights reserved.
#

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr

from phydrax.constraints import (
    AveragePressureBoundaryConstraint,
    CFDBoundaryFlowRateConstraint,
    CFDKineticEnergyFluxBoundaryConstraint,
    ContinuousIntegralBoundaryConstraint,
    ContinuousIntegralInitialConstraint,
    ContinuousIntegralInteriorConstraint,
    ContinuousODEConstraint,
    DiscreteODEConstraint,
    DiscreteTimeDataConstraint,
    EMBoundaryChargeConstraint,
    EMPoyntingFluxBoundaryConstraint,
    InitialODEConstraint,
    MagneticFluxZeroConstraint,
    SolidTotalReactionBoundaryConstraint,
)
from phydrax.domain import Cube, Interval1d, ProductStructure, TimeInterval
from phydrax.operators.differential import div, dt


def _jit_loss(constraint, functions):
    loss_fn = eqx.filter_jit(lambda k: constraint.loss(functions, key=k))
    return loss_fn(jr.key(0))


def test_continuous_ode_constraint_zero():
    time = TimeInterval(0.0, 1.0)

    @time.Function("t")
    def u(t):
        return t**2

    @time.Function("t")
    def target(t):
        return 2.0 * t

    def operator(f):
        return dt(f, var="t") - target

    structure = ProductStructure((("t",),))
    constraint = ContinuousODEConstraint(
        "u",
        time,
        operator,
        num_points=64,
        structure=structure,
    )
    loss = _jit_loss(constraint, {"u": u})
    assert loss < 1e-6


def test_discrete_ode_constraint_zero():
    time = TimeInterval(0.0, 1.0)

    @time.Function("t")
    def u(t):
        return t**2

    @time.Function("t")
    def target(t):
        return 2.0 * t

    def operator(f):
        return dt(f, var="t") - target

    times = jnp.linspace(0.0, 1.0, 8)
    constraint = DiscreteODEConstraint("u", time, operator, times=times)
    loss = _jit_loss(constraint, {"u": u})
    assert loss < 1e-6


def test_initial_ode_constraints_zero():
    time = TimeInterval(0.0, 1.0)

    @time.Function("t")
    def u(t):
        return t**2

    c0 = InitialODEConstraint("u", time, func=0.0, time_derivative_order=0)
    c1 = InitialODEConstraint("u", time, func=0.0, time_derivative_order=1)
    c2 = InitialODEConstraint(
        "u",
        time,
        func=2.0,
        time_derivative_order=2,
        time_derivative_backend="jet",
    )
    assert _jit_loss(c0, {"u": u}) < 1e-6
    assert _jit_loss(c1, {"u": u}) < 1e-6
    assert _jit_loss(c2, {"u": u}) < 1e-6


def test_discrete_time_data_constraint_zero():
    time = TimeInterval(0.0, 1.0)

    @time.Function("t")
    def u(t):
        return t**2

    times = jnp.linspace(0.0, 1.0, 6)
    values = times**2
    constraint = DiscreteTimeDataConstraint("u", time, times=times, values=values)
    loss = _jit_loss(constraint, {"u": u})
    assert loss < 1e-6


def test_integral_constraints_1d_zero_loss():
    geom = Interval1d(0.0, 1.0)
    structure = ProductStructure((("x",),))

    @geom.Function("x")
    def u(x):
        return 1.0

    @geom.Function("x")
    def v(x):
        return jnp.array([0.0])

    @geom.Function("x")
    def p(x):
        return 0.0

    @geom.Function("x")
    def D(x):
        return jnp.array([0.0])

    @geom.Function("x")
    def B(x):
        return jnp.array([0.0])

    functions = {"u": u, "v": v, "p": p, "D": D, "B": B}

    constraints = [
        ContinuousIntegralInteriorConstraint(
            "u",
            geom,
            lambda f: f,
            num_points=32,
            structure=structure,
            equal_to=1.0,
        ),
        ContinuousIntegralBoundaryConstraint(
            "u",
            geom,
            lambda f, n: f,
            num_points=8,
            structure=structure,
            equal_to=2.0,
        ),
        ContinuousIntegralInteriorConstraint(
            "v",
            geom,
            lambda f: div(f, var="x"),
            num_points=32,
            structure=structure,
        ),
        EMBoundaryChargeConstraint(
            "D",
            geom,
            total_free_charge=0.0,
            num_points=8,
            structure=structure,
        ),
        MagneticFluxZeroConstraint(
            "B",
            geom,
            num_points=8,
            structure=structure,
        ),
        CFDBoundaryFlowRateConstraint(
            "v",
            geom,
            flow_rate=0.0,
            num_points=8,
            structure=structure,
        ),
        CFDKineticEnergyFluxBoundaryConstraint(
            "v",
            geom,
            target_total_power=0.0,
            num_points=8,
            structure=structure,
        ),
        SolidTotalReactionBoundaryConstraint(
            "v",
            geom,
            lambda_=1.0,
            mu=1.0,
            target_reaction=jnp.array([0.0]),
            num_points=8,
            structure=structure,
        ),
        AveragePressureBoundaryConstraint(
            "p",
            geom,
            mean_pressure=0.0,
            num_points=8,
            structure=structure,
        ),
    ]

    for constraint in constraints:
        assert _jit_loss(constraint, functions) < 1e-6


def test_integral_initial_constraint_zero():
    geom = Interval1d(0.0, 1.0)
    time = TimeInterval(0.0, 1.0)
    domain = geom @ time
    structure = ProductStructure((("x",),))

    @domain.Function("x", "t")
    def u(x, t):
        return 1.0

    constraint = ContinuousIntegralInitialConstraint(
        "u",
        domain,
        lambda f: f,
        num_points=32,
        structure=structure,
        equal_to=1.0,
    )
    loss = _jit_loss(constraint, {"u": u})
    assert loss < 1e-6


def test_poynting_flux_constraint_zero():
    geom = Cube(center=(0.0, 0.0, 0.0), side=2.0)
    structure = ProductStructure((("x",),))

    @geom.Function("x")
    def E(x):
        return jnp.array([0.0, 0.0, 0.0])

    @geom.Function("x")
    def H(x):
        return jnp.array([0.0, 0.0, 0.0])

    constraint = EMPoyntingFluxBoundaryConstraint(
        "E",
        "H",
        geom,
        target_total_power=0.0,
        num_points=6,
        structure=structure,
    )
    loss = _jit_loss(constraint, {"E": E, "H": H})
    assert loss < 1e-6
